"""
Function module
"""

import logging
from abc import abstractmethod, ABCMeta
from collections import namedtuple
from enum import Enum
from itertools import groupby
from typing import Any, TYPE_CHECKING, Optional, Union
from collections.abc import Callable

from slither.core.cfg.scope import Scope
from slither.core.declarations.solidity_variables import (
    SolidityFunction,
    SolidityVariable,
    SolidityVariableComposed,
)
from slither.core.expressions import (
    Identifier,
    IndexAccess,
    MemberAccess,
    UnaryOperation,
)
from slither.core.solidity_types.type import Type
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.state_variable import StateVariable
from slither.utils.type import convert_type_for_solidity_signature_to_string
from slither.utils.utils import unroll


if TYPE_CHECKING:
    from slither.core.declarations import Contract, FunctionContract
    from slither.core.cfg.node import Node, NodeType
    from slither.core.variables.variable import Variable
    from slither.slithir.variables.variable import SlithIRVariable
    from slither.slithir.variables import LocalIRVariable
    from slither.core.expressions.expression import Expression
    from slither.slithir.operations import (
        HighLevelCall,
        InternalCall,
        LibraryCall,
        LowLevelCall,
        SolidityCall,
        Operation,
    )
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.core.scope.scope import FileScope

LOGGER = logging.getLogger("Function")
ReacheableNode = namedtuple("ReacheableNode", ["node", "ir"])


class ModifierStatements:
    def __init__(
        self,
        modifier: Union["Contract", "Function"],
        entry_point: "Node",
        nodes: list["Node"],
    ) -> None:
        self._modifier = modifier
        self._entry_point = entry_point
        self._nodes = nodes

    @property
    def modifier(self) -> Union["Contract", "Function"]:
        return self._modifier

    @property
    def entry_point(self) -> "Node":
        return self._entry_point

    @entry_point.setter
    def entry_point(self, entry_point: "Node"):
        self._entry_point = entry_point

    @property
    def nodes(self) -> list["Node"]:
        return self._nodes

    @nodes.setter
    def nodes(self, nodes: list["Node"]):
        self._nodes = nodes


class FunctionType(Enum):
    NORMAL = 0
    CONSTRUCTOR = 1
    FALLBACK = 2
    RECEIVE = 3
    CONSTRUCTOR_VARIABLES = 10  # Fake function to hold variable declaration statements
    CONSTRUCTOR_CONSTANT_VARIABLES = 11  # Fake function to hold variable declaration statements


def _filter_state_variables_written(expressions: list["Expression"]):
    ret = []

    for expression in expressions:
        if isinstance(expression, (Identifier, UnaryOperation, MemberAccess)):
            ret.append(expression.expression)
        elif isinstance(expression, IndexAccess):
            ret.append(expression.expression_left)
    return ret


class FunctionLanguage(Enum):
    Solidity = 0
    Yul = 1
    Vyper = 2


class Function(SourceMapping, metaclass=ABCMeta):
    """
    Function class
    """

    def __init__(self, compilation_unit: "SlitherCompilationUnit") -> None:
        super().__init__()
        self._internal_scope: list[str] = []
        self._name: str | None = None
        self._view: bool = False
        self._pure: bool = False
        self._payable: bool = False
        self._visibility: str | None = None
        self._virtual: bool = False
        self._overrides: list[FunctionContract] = []
        self._overridden_by: list[FunctionContract] = []

        self._is_implemented: bool | None = None
        self._is_empty: bool | None = None
        self._entry_point: Node | None = None
        self._nodes: list[Node] = []
        self._variables: dict[str, LocalVariable] = {}
        # slithir Temporary and references variables (but not SSA)
        self._slithir_variables: set[SlithIRVariable] = set()
        self._parameters: list[LocalVariable] = []
        self._parameters_ssa: list[LocalIRVariable] = []
        self._parameters_src: SourceMapping = SourceMapping()
        # This is used for vyper calls with default arguments
        self._default_args_as_expressions: list[Expression] = []
        self._returns: list[LocalVariable] = []
        self._returns_ssa: list[LocalIRVariable] = []
        self._returns_src: SourceMapping = SourceMapping()
        self._return_values: list[SlithIRVariable] | None = None
        self._return_values_ssa: list[SlithIRVariable] | None = None
        self._vars_read: list[Variable] = []
        self._vars_written: list[Variable] = []
        self._state_vars_read: list[StateVariable] = []
        self._vars_read_or_written: list[Variable] = []
        self._solidity_vars_read: list[SolidityVariable] = []
        self._state_vars_written: list[StateVariable] = []
        self._internal_calls: list[InternalCall] = []
        self._solidity_calls: list[SolidityCall] = []
        self._low_level_calls: list[LowLevelCall] = []
        self._high_level_calls: list[tuple[Contract, HighLevelCall]] = []
        self._library_calls: list[LibraryCall] = []
        self._external_calls_as_expressions: list[Expression] = []
        self._expression_vars_read: list[Expression] = []
        self._expression_vars_written: list[Expression] = []
        self._expression_calls: list[Expression] = []
        # self._expression_modifiers: List["Expression"] = []
        self._modifiers: list[ModifierStatements] = []
        self._explicit_base_constructor_calls: list[ModifierStatements] = []
        self._contains_assembly: bool = False

        self._expressions: list[Expression] | None = None
        self._slithir_operations: list[Operation] | None = None
        self._slithir_ssa_operations: list[Operation] | None = None

        self._all_expressions: list[Expression] | None = None
        self._all_slithir_operations: list[Operation] | None = None
        self._all_internals_calls: list[InternalCall] | None = None
        self._all_high_level_calls: list[tuple[Contract, HighLevelCall]] | None = None
        self._all_library_calls: list[LibraryCall] | None = None
        self._all_low_level_calls: list[LowLevelCall] | None = None
        self._all_solidity_calls: list[SolidityCall] | None = None
        self._all_variables_read: list[Variable] | None = None
        self._all_variables_written: list[Variable] | None = None
        self._all_state_variables_read: list[StateVariable] | None = None
        self._all_solidity_variables_read: list[SolidityVariable] | None = None
        self._all_state_variables_written: list[StateVariable] | None = None
        self._all_slithir_variables: list[SlithIRVariable] | None = None
        self._all_nodes: list[Node] | None = None
        self._all_conditional_state_variables_read: list[StateVariable] | None = None
        self._all_conditional_state_variables_read_with_loop: list[StateVariable] | None = None
        self._all_conditional_solidity_variables_read: list[SolidityVariable] | None = None
        self._all_conditional_solidity_variables_read_with_loop: list[SolidityVariable] | None = (
            None
        )
        self._all_solidity_variables_used_as_args: list[SolidityVariable] | None = None

        self._is_shadowed: bool = False
        self._shadows: bool = False

        # set(ReacheableNode)
        self._reachable_from_nodes: set[ReacheableNode] = set()
        self._reachable_from_functions: set[Function] = set()
        self._all_reachable_from_functions: set[Function] | None = None

        # Constructor, fallback, State variable constructor
        self._function_type: FunctionType | None = None
        self._is_constructor: bool | None = None

        # Computed on the fly, can be True of False
        self._can_reenter: bool | None = None
        self._can_send_eth: bool | None = None

        self._nodes_ordered_dominators: list[Node] | None = None

        self._counter_nodes = 0

        # Memoize parameters:
        # TODO: identify all the memoize parameters and add a way to undo the memoization
        self._full_name: str | None = None
        self._signature: tuple[str, list[str], list[str]] | None = None
        self._solidity_signature: str | None = None
        self._signature_str: str | None = None
        self._canonical_name: str | None = None
        self._is_protected: bool | None = None

        self.compilation_unit: SlitherCompilationUnit = compilation_unit

        self.function_language: FunctionLanguage = (
            FunctionLanguage.Solidity if compilation_unit.is_solidity else FunctionLanguage.Vyper
        )

        self._id: str | None = None

        # To be improved with a parsing of the documentation
        self.has_documentation: bool = False

    ###################################################################################
    ###################################################################################
    # region General properties
    ###################################################################################
    ###################################################################################

    @property
    def name(self) -> str:
        """
        str: function name
        """
        if self._name == "" and self._function_type == FunctionType.CONSTRUCTOR:
            return "constructor"
        if self._name == "" and self._function_type == FunctionType.FALLBACK:
            return "fallback"
        if self._function_type == FunctionType.RECEIVE:
            return "receive"
        if self._function_type == FunctionType.CONSTRUCTOR_VARIABLES:
            return "slitherConstructorVariables"
        if self._function_type == FunctionType.CONSTRUCTOR_CONSTANT_VARIABLES:
            return "slitherConstructorConstantVariables"
        return self._name

    @name.setter
    def name(self, new_name: str):
        self._name = new_name

    @property
    def internal_scope(self) -> list[str]:
        """
        Return a list of name representing the scope of the function
        This is used to model nested functions declared in YUL

        :return:
        """
        return self._internal_scope

    @internal_scope.setter
    def internal_scope(self, new_scope: list[str]):
        self._internal_scope = new_scope

    @property
    def full_name(self) -> str:
        """
        str: func_name(type1,type2)
        Return the function signature without the return values
        The difference between this function and solidity_function is that full_name does not translate the underlying
        type (ex: structure, contract to address, ...)
        """
        if self._full_name is None:
            name, parameters, _ = self.signature
            full_name = ".".join(self._internal_scope + [name]) + "(" + ",".join(parameters) + ")"
            self._full_name = full_name
        return self._full_name

    @property
    @abstractmethod
    def canonical_name(self) -> str:
        """
        str: contract.func_name(type1,type2)
        Return the function signature without the return values
        """
        return ""

    @property
    def contains_assembly(self) -> bool:
        return self._contains_assembly

    @contains_assembly.setter
    def contains_assembly(self, c: bool):
        self._contains_assembly = c

    def can_reenter(self, callstack: list[Union["Function", "Variable"]] | None = None) -> bool:
        """
        Check if the function can re-enter
        Follow internal calls.
        Do not consider CREATE as potential re-enter, but check if the
        destination's constructor can contain a call (recurs. follow nested CREATE)
        For Solidity > 0.5, filter access to public variables and constant/pure/view
        For call to this. check if the destination can re-enter
        Do not consider Send/Transfer as there is not enough gas
        :param callstack: used internally to check for recursion
        :return bool:
        """
        from slither.slithir.operations import Call

        if self._can_reenter is None:
            self._can_reenter = False
            for ir in self.all_slithir_operations():
                if isinstance(ir, Call) and ir.can_reenter(callstack):
                    self._can_reenter = True
                    return True
        return self._can_reenter

    def can_send_eth(self) -> bool:
        """
        Check if the function or any internal (not external) functions called by it can send eth
        :return bool:
        """
        from slither.slithir.operations import Call

        if self._can_send_eth is None:
            self._can_send_eth = False
            for ir in self.all_slithir_operations():
                if isinstance(ir, Call) and ir.can_send_eth():
                    self._can_send_eth = True
                    return True
        return self._can_send_eth

    @property
    def is_checked(self) -> bool:
        """
        Return true if the overflow are enabled by default


        :return:
        """

        return self.compilation_unit.solc_version >= "0.8.0"

    @property
    def id(self) -> str | None:
        """
        Return the reference ID of the function, if available.

        :return:
        :rtype:
        """
        return self._id

    @id.setter
    def id(self, new_id: str):
        self._id = new_id

    @property
    @abstractmethod
    def file_scope(self) -> "FileScope":
        pass

    # endregion
    ###################################################################################
    ###################################################################################
    # region Type (FunctionType)
    ###################################################################################
    ###################################################################################

    def set_function_type(self, t: FunctionType) -> None:
        assert isinstance(t, FunctionType)
        self._function_type = t

    @property
    def function_type(self) -> FunctionType | None:
        return self._function_type

    @function_type.setter
    def function_type(self, t: FunctionType):
        self._function_type = t

    @property
    def is_constructor(self) -> bool:
        """
        bool: True if the function is the constructor
        """
        return self._function_type == FunctionType.CONSTRUCTOR

    @property
    def is_constructor_variables(self) -> bool:
        """
        bool: True if the function is the constructor of the variables
        Slither has inbuilt functions to hold the state variables initialization
        """
        return self._function_type in [
            FunctionType.CONSTRUCTOR_VARIABLES,
            FunctionType.CONSTRUCTOR_CONSTANT_VARIABLES,
        ]

    @property
    def is_fallback(self) -> bool:
        """
            Determine if the function is the fallback function for the contract
        Returns
            (bool)
        """
        return self._function_type == FunctionType.FALLBACK

    @property
    def is_receive(self) -> bool:
        """
            Determine if the function is the receive function for the contract
        Returns
            (bool)
        """
        return self._function_type == FunctionType.RECEIVE

    # endregion
    ###################################################################################
    ###################################################################################
    # region Payable
    ###################################################################################
    ###################################################################################

    @property
    def payable(self) -> bool:
        """
        bool: True if the function is payable
        """
        return self._payable

    @payable.setter
    def payable(self, p: bool):
        self._payable = p

    # endregion
    ###################################################################################
    ###################################################################################
    # region Virtual
    ###################################################################################
    ###################################################################################

    @property
    def is_virtual(self) -> bool:
        """
        Note for Solidity < 0.6.0 it will always be false
        bool: True if the function is virtual
        """
        return self._virtual

    @is_virtual.setter
    def is_virtual(self, v: bool):
        self._virtual = v

    @property
    def is_override(self) -> bool:
        """
        Note for Solidity < 0.6.0 it will always be false
        bool: True if the function overrides a base function
        """
        return len(self._overrides) > 0

    @property
    def overridden_by(self) -> list["FunctionContract"]:
        """
        List["FunctionContract"]: List of functions in child contracts that override this function
        This may include distinct instances of the same function due to inheritance
        """
        return self._overridden_by

    @property
    def overrides(self) -> list["FunctionContract"]:
        """
        List["FunctionContract"]: List of functions in parent contracts that this function overrides
        This may include distinct instances of the same function due to inheritance
        """
        return self._overrides

    # endregion
    ###################################################################################
    ###################################################################################
    # region Visibility
    ###################################################################################
    ###################################################################################

    @property
    def visibility(self) -> str:
        """
        str: Function visibility
        """
        assert self._visibility is not None
        return self._visibility

    @visibility.setter
    def visibility(self, v: str):
        self._visibility = v

    def set_visibility(self, v: str) -> None:
        self._visibility = v

    @property
    def view(self) -> bool:
        """
        bool: True if the function is declared as view
        """
        return self._view

    @view.setter
    def view(self, v: bool):
        self._view = v

    @property
    def pure(self) -> bool:
        """
        bool: True if the function is declared as pure
        """
        return self._pure

    @pure.setter
    def pure(self, p: bool):
        self._pure = p

    @property
    def is_shadowed(self) -> bool:
        return self._is_shadowed

    @is_shadowed.setter
    def is_shadowed(self, is_shadowed):
        self._is_shadowed = is_shadowed

    @property
    def shadows(self) -> bool:
        return self._shadows

    @shadows.setter
    def shadows(self, _shadows: bool):
        self._shadows = _shadows

    # endregion
    ###################################################################################
    ###################################################################################
    # region Function's body
    ###################################################################################
    ###################################################################################

    @property
    def is_implemented(self) -> bool:
        """
        bool: True if the function is implemented
        """
        return self._is_implemented

    @is_implemented.setter
    def is_implemented(self, is_impl: bool):
        self._is_implemented = is_impl

    @property
    def is_empty(self) -> bool:
        """
        bool: True if the function is empty, None if the function is an interface
        """
        return self._is_empty

    @is_empty.setter
    def is_empty(self, empty: bool):
        self._is_empty = empty

    # endregion
    ###################################################################################
    ###################################################################################
    # region Nodes
    ###################################################################################
    ###################################################################################

    @property
    def nodes(self) -> list["Node"]:
        """
        list(Node): List of the nodes
        """
        return list(self._nodes)

    @nodes.setter
    def nodes(self, nodes: list["Node"]):
        self._nodes = nodes

    @property
    def entry_point(self) -> Optional["Node"]:
        """
        Node: Entry point of the function
        """
        return self._entry_point

    @entry_point.setter
    def entry_point(self, node: "Node"):
        self._entry_point = node

    def add_node(self, node: "Node") -> None:
        if not self._entry_point:
            self._entry_point = node
        self._nodes.append(node)

    @property
    def nodes_ordered_dominators(self) -> list["Node"]:
        # TODO: does not work properly; most likely due to modifier call
        # This will not work for modifier call that lead to multiple nodes
        # from slither.core.cfg.node import NodeType
        if self._nodes_ordered_dominators is None:
            self._nodes_ordered_dominators = []
            if self.entry_point:
                self._compute_nodes_ordered_dominators(self.entry_point)

            for node in self.nodes:
                # if node.type == NodeType.OTHER_ENTRYPOINT:
                if node not in self._nodes_ordered_dominators:
                    self._compute_nodes_ordered_dominators(node)

        return self._nodes_ordered_dominators

    def _compute_nodes_ordered_dominators(self, node: "Node"):
        assert self._nodes_ordered_dominators is not None
        if node in self._nodes_ordered_dominators:
            return
        self._nodes_ordered_dominators.append(node)
        for dom in node.dominance_exploration_ordered:
            self._compute_nodes_ordered_dominators(dom)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Parameters
    ###################################################################################
    ###################################################################################

    @property
    def parameters(self) -> list["LocalVariable"]:
        """
        list(LocalVariable): List of the parameters
        """
        return list(self._parameters)

    def add_parameters(self, p: "LocalVariable") -> None:
        self._parameters.append(p)

    @property
    def parameters_ssa(self) -> list["LocalIRVariable"]:
        """
        list(LocalIRVariable): List of the parameters (SSA form)
        """
        return list(self._parameters_ssa)

    def add_parameter_ssa(self, var: "LocalIRVariable") -> None:
        self._parameters_ssa.append(var)

    def parameters_src(self) -> SourceMapping:
        return self._parameters_src

    # endregion
    ###################################################################################
    ###################################################################################
    # region Return values
    ###################################################################################
    ###################################################################################

    @property
    def return_type(self) -> list[Type] | None:
        """
        Return the list of return type
        If no return, return None
        """
        returns = self.returns
        if returns:
            return [r.type for r in returns]
        return None

    def returns_src(self) -> SourceMapping:
        return self._returns_src

    @property
    def type(self) -> list[Type] | None:
        """
        Return the list of return type
        If no return, return None
        Alias of return_type
        """
        return self.return_type

    @property
    def returns(self) -> list["LocalVariable"]:
        """
        list(LocalVariable): List of the return variables
        """
        return list(self._returns)

    def add_return(self, r: "LocalVariable") -> None:
        self._returns.append(r)

    @property
    def returns_ssa(self) -> list["LocalIRVariable"]:
        """
        list(LocalIRVariable): List of the return variables (SSA form)
        """
        return list(self._returns_ssa)

    def add_return_ssa(self, var: "LocalIRVariable") -> None:
        self._returns_ssa.append(var)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Modifiers
    ###################################################################################
    ###################################################################################

    @property
    def modifiers(self) -> list[Union["Contract", "Function"]]:
        """
        list(Modifier): List of the modifiers
        Can be contract for constructor's calls

        """
        return [c.modifier for c in self._modifiers]

    def add_modifier(self, modif: "ModifierStatements") -> None:
        self._modifiers.append(modif)

    @property
    def modifiers_statements(self) -> list[ModifierStatements]:
        """
        list(ModifierCall): List of the modifiers call (include expression and irs)
        """
        return list(self._modifiers)

    @property
    def explicit_base_constructor_calls(self) -> list["Function"]:
        """
        list(Function): List of the base constructors called explicitly by this presumed constructor definition.

                        Base constructors implicitly or explicitly called by the contract definition will not be
                        included.
        """
        # This is a list of contracts internally, so we convert it to a list of constructor functions.
        return [
            c.modifier.constructors_declared
            for c in self._explicit_base_constructor_calls
            if c.modifier.constructors_declared
        ]

    @property
    def explicit_base_constructor_calls_statements(self) -> list[ModifierStatements]:
        """
        list(ModifierCall): List of the base constructors called explicitly by this presumed constructor definition.

        """
        # This is a list of contracts internally, so we convert it to a list of constructor functions.
        return list(self._explicit_base_constructor_calls)

    def add_explicit_base_constructor_calls_statements(self, modif: ModifierStatements) -> None:
        self._explicit_base_constructor_calls.append(modif)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Variables
    ###################################################################################
    ###################################################################################

    @property
    def variables(self) -> list[LocalVariable]:
        """
        Return all local variables
        Include parameters and return values
        """
        return list(self._variables.values())

    @property
    def local_variables(self) -> list[LocalVariable]:
        """
        Return all local variables (dont include parameters and return values)
        """
        return list(set(self.variables) - set(self.returns) - set(self.parameters))

    @property
    def variables_as_dict(self) -> dict[str, LocalVariable]:
        return self._variables

    @property
    def variables_read(self) -> list["Variable"]:
        """
        list(Variable): Variables read (local/state/solidity)
        """
        return list(self._vars_read)

    @property
    def variables_written(self) -> list["Variable"]:
        """
        list(Variable): Variables written (local/state/solidity)
        """
        return list(self._vars_written)

    @property
    def state_variables_read(self) -> list["StateVariable"]:
        """
        list(StateVariable): State variables read
        """
        return list(self._state_vars_read)

    @property
    def solidity_variables_read(self) -> list["SolidityVariable"]:
        """
        list(SolidityVariable): Solidity variables read
        """
        return list(self._solidity_vars_read)

    @property
    def state_variables_written(self) -> list["StateVariable"]:
        """
        list(StateVariable): State variables written
        """
        return list(self._state_vars_written)

    @property
    def variables_read_or_written(self) -> list["Variable"]:
        """
        list(Variable): Variables read or written (local/state/solidity)
        """
        return list(self._vars_read_or_written)

    @property
    def variables_read_as_expression(self) -> list["Expression"]:
        return self._expression_vars_read

    @property
    def variables_written_as_expression(self) -> list["Expression"]:
        return self._expression_vars_written

    @property
    def slithir_variables(self) -> list["SlithIRVariable"]:
        """
        Temporary and Reference Variables (not SSA form)
        """

        return list(self._slithir_variables)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Calls
    ###################################################################################
    ###################################################################################

    @property
    def internal_calls(self) -> list["InternalCall"]:
        """
        list(InternalCall): List of IR operations for internal calls
        """
        return list(self._internal_calls)

    @property
    def solidity_calls(self) -> list["SolidityCall"]:
        """
        list(SolidityCall): List of IR operations for Solidity calls
        """
        return list(self._solidity_calls)

    @property
    def high_level_calls(self) -> list[tuple["Contract", "HighLevelCall"]]:
        """
        list(Tuple(Contract, "HighLevelCall")): List of call target contract and IR of the high level call
        A variable is called in case of call to a public state variable
        Include library calls
        """
        return list(self._high_level_calls)

    @property
    def library_calls(self) -> list["LibraryCall"]:
        """
        list(LibraryCall): List of IR operations for library calls
        """
        return list(self._library_calls)

    @property
    def low_level_calls(self) -> list["LowLevelCall"]:
        """
        list(LowLevelCall): List of IR operations for low level calls
        A low level call is defined by
        - the variable called
        - the name of the function (call/delegatecall/callcode)
        """
        return list(self._low_level_calls)

    @property
    def external_calls_as_expressions(self) -> list["Expression"]:
        """
        list(ExpressionCall): List of message calls (that creates a transaction)
        """
        return list(self._external_calls_as_expressions)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Expressions
    ###################################################################################
    ###################################################################################

    @property
    def calls_as_expressions(self) -> list["Expression"]:
        return self._expression_calls

    @property
    def expressions(self) -> list["Expression"]:
        """
        list(Expression): List of the expressions
        """
        if self._expressions is None:
            expressionss = [n.expression for n in self.nodes]
            expressions = [e for e in expressionss if e]
            self._expressions = expressions
        return self._expressions

    @property
    def return_values(self) -> list["SlithIRVariable"]:
        """
        list(Return Values): List of the return values
        """
        from slither.core.cfg.node import NodeType
        from slither.slithir.operations import Return
        from slither.slithir.variables import Constant

        if self._return_values is None:
            return_values = []
            returns = [n for n in self.nodes if n.type == NodeType.RETURN]
            [
                return_values.extend(ir.values)
                for node in returns
                for ir in node.irs
                if isinstance(ir, Return)
            ]
            self._return_values = list({x for x in return_values if not isinstance(x, Constant)})
        return self._return_values

    @property
    def return_values_ssa(self) -> list["SlithIRVariable"]:
        """
        list(Return Values in SSA form): List of the return values in ssa form
        """
        from slither.core.cfg.node import NodeType
        from slither.slithir.operations import Return
        from slither.slithir.variables import Constant

        if self._return_values_ssa is None:
            return_values_ssa = []
            returns = [n for n in self.nodes if n.type == NodeType.RETURN]
            [
                return_values_ssa.extend(ir.values)
                for node in returns
                for ir in node.irs_ssa
                if isinstance(ir, Return)
            ]
            self._return_values_ssa = list(
                {x for x in return_values_ssa if not isinstance(x, Constant)}
            )
        return self._return_values_ssa

    # endregion
    ###################################################################################
    ###################################################################################
    # region SlithIR
    ###################################################################################
    ###################################################################################

    @property
    def slithir_operations(self) -> list["Operation"]:
        """
        list(Operation): List of the slithir operations
        """
        if self._slithir_operations is None:
            operationss = [n.irs for n in self.nodes]
            operations = [item for sublist in operationss for item in sublist if item]
            self._slithir_operations = operations
        return self._slithir_operations

    @property
    def slithir_ssa_operations(self) -> list["Operation"]:
        """
        list(Operation): List of the slithir operations (SSA)
        """
        if self._slithir_ssa_operations is None:
            operationss = [n.irs_ssa for n in self.nodes]
            operations = [item for sublist in operationss for item in sublist if item]
            self._slithir_ssa_operations = operations
        return self._slithir_ssa_operations

    # endregion
    ###################################################################################
    ###################################################################################
    # region Signature
    ###################################################################################
    ###################################################################################

    @property
    def solidity_signature(self) -> str:
        """
        Return a signature following the Solidity Standard
        Contract and converted into address

        It might still keep internal types (ex: structure name) for internal functions.
        The reason is that internal functions allows recursive structure definition, which
        can't be converted following the Solidity stand ard

        :return: the solidity signature
        """
        if self._solidity_signature is None:
            parameters = [
                convert_type_for_solidity_signature_to_string(x.type) for x in self.parameters
            ]
            self._solidity_signature = self.name + "(" + ",".join(parameters) + ")"
        return self._solidity_signature

    @property
    def signature(self) -> tuple[str, list[str], list[str]]:
        """
        (str, list(str), list(str)): Function signature as
        (name, list parameters type, list return values type)
        """
        # FIXME memoizing this function is not working properly for vyper
        # if self._signature is None:
        return (
            self.name,
            [str(x.type) for x in self.parameters],
            [str(x.type) for x in self.returns],
        )
        #     self._signature = signature
        # return self._signature

    @property
    def signature_str(self) -> str:
        """
        str: func_name(type1,type2) returns (type3)
        Return the function signature as a str (contains the return values)
        """
        if self._signature_str is None:
            name, parameters, returnVars = self.signature
            self._signature_str = (
                name + "(" + ",".join(parameters) + ") returns(" + ",".join(returnVars) + ")"
            )
        return self._signature_str

    # endregion
    ###################################################################################
    ###################################################################################
    # region Functions
    ###################################################################################
    ###################################################################################

    @property
    @abstractmethod
    def functions_shadowed(self) -> list["Function"]:
        pass

    # endregion
    ###################################################################################
    ###################################################################################
    # region Reachable
    ###################################################################################
    ###################################################################################

    @property
    def reachable_from_nodes(self) -> set[ReacheableNode]:
        """
        Return
            ReacheableNode
        """
        return self._reachable_from_nodes

    @property
    def reachable_from_functions(self) -> set["Function"]:
        return self._reachable_from_functions

    @property
    def all_reachable_from_functions(self) -> set["Function"]:
        """
        Give the recursive version of reachable_from_functions (all the functions that lead to call self in the CFG)
        """
        if self._all_reachable_from_functions is None:
            functions: set[Function] = set()

            new_functions = self.reachable_from_functions
            # iterate until we have are finding new functions
            while new_functions and not new_functions.issubset(functions):
                functions = functions.union(new_functions)
                # Use a temporary set, because we iterate over new_functions
                new_functionss: set[Function] = set()
                for f in new_functions:
                    new_functionss = new_functionss.union(f.reachable_from_functions)
                new_functions = new_functionss - functions

            self._all_reachable_from_functions = functions
        return self._all_reachable_from_functions

    def add_reachable_from_node(self, n: "Node", ir: "Operation") -> None:
        self._reachable_from_nodes.add(ReacheableNode(n, ir))
        self._reachable_from_functions.add(n.function)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Recursive getters
    ###################################################################################
    ###################################################################################

    def _explore_functions(self, f_new_values: Callable[["Function"], list]) -> list[Any]:
        values = f_new_values(self)
        explored = [self]
        to_explore = [
            ir.function
            for ir in self.internal_calls
            if isinstance(ir.function, Function) and ir.function not in explored
        ]
        to_explore += [
            ir.function
            for ir in self.library_calls
            if isinstance(ir.function, Function) and ir.function not in explored
        ]
        to_explore += [m for m in self.modifiers if m not in explored]

        while to_explore:
            f = to_explore[0]
            to_explore = to_explore[1:]
            if f in explored:
                continue
            explored.append(f)

            values += f_new_values(f)

            to_explore += [
                ir.function
                for ir in f.internal_calls
                if isinstance(ir.function, Function)
                and ir.function not in explored
                and ir.function not in to_explore
            ]
            to_explore += [
                ir.function
                for ir in f.library_calls
                if isinstance(ir.function, Function)
                and ir.function not in explored
                and ir.function not in to_explore
            ]
            to_explore += [m for m in f.modifiers if m not in explored and m not in to_explore]

        return list(set(values))

    def all_variables_read(self) -> list["Variable"]:
        """recursive version of variables_read"""
        if self._all_variables_read is None:
            self._all_variables_read = self._explore_functions(lambda x: x.variables_read)
        return self._all_variables_read

    def all_variables_written(self) -> list["Variable"]:
        """recursive version of variables_written"""
        if self._all_variables_written is None:
            self._all_variables_written = self._explore_functions(lambda x: x.variables_written)
        return self._all_variables_written

    def all_state_variables_read(self) -> list["StateVariable"]:
        """recursive version of variables_read"""
        if self._all_state_variables_read is None:
            self._all_state_variables_read = self._explore_functions(
                lambda x: x.state_variables_read
            )
        return self._all_state_variables_read

    def all_solidity_variables_read(self) -> list[SolidityVariable]:
        """recursive version of solidity_read"""
        if self._all_solidity_variables_read is None:
            self._all_solidity_variables_read = self._explore_functions(
                lambda x: x.solidity_variables_read
            )
        return self._all_solidity_variables_read

    def all_slithir_variables(self) -> list["SlithIRVariable"]:
        """recursive version of slithir_variables"""
        if self._all_slithir_variables is None:
            self._all_slithir_variables = self._explore_functions(lambda x: x.slithir_variables)
        return self._all_slithir_variables

    def all_nodes(self) -> list["Node"]:
        """recursive version of nodes"""
        if self._all_nodes is None:
            self._all_nodes = self._explore_functions(lambda x: x.nodes)
        return self._all_nodes

    def all_expressions(self) -> list["Expression"]:
        """recursive version of variables_read"""
        if self._all_expressions is None:
            self._all_expressions = self._explore_functions(lambda x: x.expressions)
        return self._all_expressions

    def all_slithir_operations(self) -> list["Operation"]:
        if self._all_slithir_operations is None:
            self._all_slithir_operations = self._explore_functions(lambda x: x.slithir_operations)
        return self._all_slithir_operations

    def all_state_variables_written(self) -> list[StateVariable]:
        """recursive version of variables_written"""
        if self._all_state_variables_written is None:
            self._all_state_variables_written = self._explore_functions(
                lambda x: x.state_variables_written
            )
        return self._all_state_variables_written

    def all_internal_calls(self) -> list["InternalCall"]:
        """recursive version of internal_calls"""
        if self._all_internals_calls is None:
            self._all_internals_calls = self._explore_functions(lambda x: x.internal_calls)
        return self._all_internals_calls

    def all_low_level_calls(self) -> list["LowLevelCall"]:
        """recursive version of low_level calls"""
        if self._all_low_level_calls is None:
            self._all_low_level_calls = self._explore_functions(lambda x: x.low_level_calls)
        return self._all_low_level_calls

    def all_high_level_calls(self) -> list[tuple["Contract", "HighLevelCall"]]:
        """recursive version of high_level calls"""
        if self._all_high_level_calls is None:
            self._all_high_level_calls = self._explore_functions(lambda x: x.high_level_calls)
        return self._all_high_level_calls

    def all_library_calls(self) -> list["LibraryCall"]:
        """recursive version of library calls"""
        if self._all_library_calls is None:
            self._all_library_calls = self._explore_functions(lambda x: x.library_calls)
        return self._all_library_calls

    def all_solidity_calls(self) -> list["SolidityCall"]:
        """recursive version of solidity calls"""
        if self._all_solidity_calls is None:
            self._all_solidity_calls = self._explore_functions(lambda x: x.solidity_calls)
        return self._all_solidity_calls

    @staticmethod
    def _explore_func_cond_read(func: "Function", include_loop: bool) -> list["StateVariable"]:
        ret = [n.state_variables_read for n in func.nodes if n.is_conditional(include_loop)]
        return [item for sublist in ret for item in sublist]

    def all_conditional_state_variables_read(self, include_loop=True) -> list["StateVariable"]:
        """
        Return the state variable used in a condition

        Over approximate and also return index access
        It won't work if the variable is assigned to a temp variable
        """
        if include_loop:
            if self._all_conditional_state_variables_read_with_loop is None:
                self._all_conditional_state_variables_read_with_loop = self._explore_functions(
                    lambda x: self._explore_func_cond_read(x, include_loop)
                )
            return self._all_conditional_state_variables_read_with_loop
        if self._all_conditional_state_variables_read is None:
            self._all_conditional_state_variables_read = self._explore_functions(
                lambda x: self._explore_func_cond_read(x, include_loop)
            )
        return self._all_conditional_state_variables_read

    @staticmethod
    def _solidity_variable_in_binary(node: "Node") -> list[SolidityVariable]:
        from slither.slithir.operations.binary import Binary

        ret = []
        for ir in node.irs:
            if isinstance(ir, Binary):
                ret += ir.read
        return [var for var in ret if isinstance(var, SolidityVariable)]

    @staticmethod
    def _explore_func_conditional(
        func: "Function",
        f: Callable[["Node"], list[SolidityVariable]],
        include_loop: bool,
    ) -> list[Any]:
        ret = [f(n) for n in func.nodes if n.is_conditional(include_loop)]
        return [item for sublist in ret for item in sublist]

    def all_conditional_solidity_variables_read(
        self, include_loop: bool = True
    ) -> list[SolidityVariable]:
        """
        Return the Solidity variables directly used in a condition

        Use of the IR to filter index access
        Assumption: the solidity vars are used directly in the conditional node
        It won't work if the variable is assigned to a temp variable
        """
        if include_loop:
            if self._all_conditional_solidity_variables_read_with_loop is None:
                self._all_conditional_solidity_variables_read_with_loop = self._explore_functions(
                    lambda x: self._explore_func_conditional(
                        x, self._solidity_variable_in_binary, include_loop
                    )
                )
            return self._all_conditional_solidity_variables_read_with_loop

        if self._all_conditional_solidity_variables_read is None:
            self._all_conditional_solidity_variables_read = self._explore_functions(
                lambda x: self._explore_func_conditional(
                    x, self._solidity_variable_in_binary, include_loop
                )
            )
        return self._all_conditional_solidity_variables_read

    @staticmethod
    def _solidity_variable_in_internal_calls(node: "Node") -> list[SolidityVariable]:
        from slither.slithir.operations.internal_call import InternalCall

        ret = []
        for ir in node.irs:
            if isinstance(ir, InternalCall):
                ret += ir.read
        return [var for var in ret if isinstance(var, SolidityVariable)]

    @staticmethod
    def _explore_func_nodes(
        func: "Function", f: Callable[["Node"], list[SolidityVariable]]
    ) -> list[Any | SolidityVariableComposed]:
        ret = [f(n) for n in func.nodes]
        return [item for sublist in ret for item in sublist]

    def all_solidity_variables_used_as_args(self) -> list[SolidityVariable]:
        """
        Return the Solidity variables directly used in a call

        Use of the IR to filter index access
        Used to catch check(msg.sender)
        """
        if self._all_solidity_variables_used_as_args is None:
            self._all_solidity_variables_used_as_args = self._explore_functions(
                lambda x: self._explore_func_nodes(x, self._solidity_variable_in_internal_calls)
            )
        return self._all_solidity_variables_used_as_args

    # endregion
    ###################################################################################
    ###################################################################################
    # region Visitor
    ###################################################################################
    ###################################################################################

    def apply_visitor(self, Visitor: Callable) -> list:
        """
            Apply a visitor to all the function expressions
        Args:
            Visitor: slither.visitors
        Returns
            list(): results of the visit
        """
        expressions = self.expressions
        v = [Visitor(e).result() for e in expressions]
        return [item for sublist in v for item in sublist]

    # endregion
    ###################################################################################
    ###################################################################################
    # region Getters from/to object
    ###################################################################################
    ###################################################################################

    def get_local_variable_from_name(self, variable_name: str) -> LocalVariable | None:
        """
            Return a local variable from a name

        Args:
            variable_name (str): name of the variable
        Returns:
            LocalVariable
        """
        return next((v for v in self.variables if v.name == variable_name), None)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Export
    ###################################################################################
    ###################################################################################

    def cfg_to_dot(self, filename: str):
        """
            Export the function to a dot file
        Args:
            filename (str)
        """
        with open(filename, "w", encoding="utf8") as f:
            f.write("digraph{\n")
            for node in self.nodes:
                f.write(f'{node.node_id}[label="{node!s}"];\n')
                for son in node.sons:
                    f.write(f"{node.node_id}->{son.node_id};\n")

            f.write("}\n")

    def dominator_tree_to_dot(self, filename: str):
        """
            Export the dominator tree of the function to a dot file
        Args:
            filename (str)
        """

        def description(node):
            desc = f"{node}\n"
            desc += f"id: {node.node_id}"
            if node.dominance_frontier:
                desc += f"\ndominance frontier: {[n.node_id for n in node.dominance_frontier]}"
            return desc

        with open(filename, "w", encoding="utf8") as f:
            f.write("digraph{\n")
            for node in self.nodes:
                f.write(f'{node.node_id}[label="{description(node)}"];\n')
                if node.immediate_dominator:
                    f.write(f"{node.immediate_dominator.node_id}->{node.node_id};\n")

            f.write("}\n")

    def slithir_cfg_to_dot(self, filename: str):
        """
        Export the CFG to a DOT file. The nodes includes the Solidity expressions and the IRs
        :param filename:
        :return:
        """
        content = self.slithir_cfg_to_dot_str()
        with open(filename, "w", encoding="utf8") as f:
            f.write(content)

    def slithir_cfg_to_dot_str(self, skip_expressions: bool = False) -> str:
        """
        Export the CFG to a DOT format. The nodes includes the Solidity expressions and the IRs
        :return: the DOT content
        :rtype: str
        """
        from slither.core.cfg.node import NodeType

        content = ""
        content += "digraph{\n"
        for node in self.nodes:
            label = f"Node Type: {node.type.value} {node.node_id}\n"
            if node.expression and not skip_expressions:
                label += f"\nEXPRESSION:\n{node.expression}\n"
            if node.irs and not skip_expressions:
                label += "\nIRs:\n" + "\n".join([str(ir) for ir in node.irs])
            content += f'{node.node_id}[label="{label}"];\n'
            if node.type in [NodeType.IF, NodeType.IFLOOP]:
                true_node = node.son_true
                if true_node:
                    content += f'{node.node_id}->{true_node.node_id}[label="True"];\n'
                false_node = node.son_false
                if false_node:
                    content += f'{node.node_id}->{false_node.node_id}[label="False"];\n'
            else:
                for son in node.sons:
                    content += f"{node.node_id}->{son.node_id};\n"

        content += "}\n"
        return content

    # endregion
    ###################################################################################
    ###################################################################################
    # region Summary information
    ###################################################################################
    ###################################################################################

    def is_reading(self, variable: "Variable") -> bool:
        """
            Check if the function reads the variable
        Args:
            variable (Variable):
        Returns:
            bool: True if the variable is read
        """
        return variable in self.variables_read

    def is_reading_in_conditional_node(self, variable: "Variable") -> bool:
        """
            Check if the function reads the variable in a IF node
        Args:
            variable (Variable):
        Returns:
            bool: True if the variable is read
        """
        variables_reads = [n.variables_read for n in self.nodes if n.contains_if()]
        variables_read = [item for sublist in variables_reads for item in sublist]
        return variable in variables_read

    def is_reading_in_require_or_assert(self, variable: "Variable") -> bool:
        """
            Check if the function reads the variable in an require or assert
        Args:
            variable (Variable):
        Returns:
            bool: True if the variable is read
        """
        variables_reads = [n.variables_read for n in self.nodes if n.contains_require_or_assert()]
        variables_read = [item for sublist in variables_reads for item in sublist]
        return variable in variables_read

    def is_writing(self, variable: "Variable") -> bool:
        """
            Check if the function writes the variable
        Args:
            variable (Variable):
        Returns:
            bool: True if the variable is written
        """
        return variable in self.variables_written

    @abstractmethod
    def get_summary(
        self,
    ) -> tuple[str, str, str, list[str], list[str], list[str], list[str], list[str]]:
        pass

    def is_protected(self) -> bool:
        """
            Determine if the function is protected using a check on msg.sender

            Consider onlyOwner as a safe modifier.
            If the owner functionality is incorrectly implemented, this will lead to incorrectly
            classify the function as protected

            Otherwise only detects if msg.sender is directly used in a condition
            For example, it wont work for:
                address a = msg.sender
                require(a == owner)
        Returns
            (bool)
        """

        if self._is_protected is None:
            if self.is_constructor:
                self._is_protected = True
                return True
            if "onlyOwner" in [m.name for m in self.modifiers]:
                self._is_protected = True
                return True
            conditional_vars = self.all_conditional_solidity_variables_read(include_loop=False)
            args_vars = self.all_solidity_variables_used_as_args()
            self._is_protected = (
                SolidityVariableComposed("msg.sender") in conditional_vars + args_vars
            )
        return self._is_protected

    @property
    def is_reentrant(self) -> bool:
        """
        Determine if the function can be re-entered
        """
        reentrancy_modifier = "nonReentrant"

        if self.function_language == FunctionLanguage.Vyper:
            reentrancy_modifier = "nonreentrant(lock)"

        # TODO: compare with hash of known nonReentrant modifier instead of the name
        if reentrancy_modifier in [m.name for m in self.modifiers]:
            return False

        if self.visibility in ["public", "external"]:
            return True

        # If it's an internal function, check if all its entry points have the nonReentrant modifier
        all_entry_points = [
            f for f in self.all_reachable_from_functions if f.visibility in ["public", "external"]
        ]
        if not all_entry_points:
            return True
        return not all(
            reentrancy_modifier in [m.name for m in f.modifiers] for f in all_entry_points
        )

    # endregion
    ###################################################################################
    ###################################################################################
    # region Analyses
    ###################################################################################
    ###################################################################################

    def _analyze_read_write(self) -> None:
        """Compute variables read/written/..."""
        write_var = [x.variables_written_as_expression for x in self.nodes]
        write_var = [x for x in write_var if x]
        write_var = [item for sublist in write_var for item in sublist]
        write_var = list(set(write_var))
        # Remove duplicate if they share the same string representation
        write_var = [
            next(obj)
            for i, obj in groupby(sorted(write_var, key=lambda x: str(x)), lambda x: str(x))
        ]
        self._expression_vars_written = write_var

        write_var = [x.variables_written for x in self.nodes]
        write_var = [x for x in write_var if x]
        write_var = [item for sublist in write_var for item in sublist]
        write_var = list(set(write_var))
        # Remove duplicate if they share the same string representation
        write_var = [
            next(obj)
            for i, obj in groupby(sorted(write_var, key=lambda x: str(x)), lambda x: str(x))
        ]
        self._vars_written = write_var

        read_var = [x.variables_read_as_expression for x in self.nodes]
        read_var = [x for x in read_var if x]
        read_var = [item for sublist in read_var for item in sublist]
        # Remove duplicate if they share the same string representation
        read_var = [
            next(obj)
            for i, obj in groupby(sorted(read_var, key=lambda x: str(x)), lambda x: str(x))
        ]
        self._expression_vars_read = read_var

        read_var = [x.variables_read for x in self.nodes]
        read_var = [x for x in read_var if x]
        read_var = [item for sublist in read_var for item in sublist]
        # Remove duplicate if they share the same string representation
        read_var = [
            next(obj)
            for i, obj in groupby(sorted(read_var, key=lambda x: str(x)), lambda x: str(x))
        ]
        self._vars_read = read_var

        self._state_vars_written = [
            x for x in self.variables_written if isinstance(x, StateVariable)
        ]
        self._state_vars_read = [x for x in self.variables_read if isinstance(x, StateVariable)]
        self._solidity_vars_read = [
            x for x in self.variables_read if isinstance(x, SolidityVariable)
        ]

        self._vars_read_or_written = self._vars_written + self._vars_read

        slithir_variables = [x.slithir_variables for x in self.nodes]
        slithir_variables = [x for x in slithir_variables if x]
        self._slithir_variables = [item for sublist in slithir_variables for item in sublist]

    def _analyze_calls(self) -> None:
        calls = [x.calls_as_expression for x in self.nodes]
        calls = [x for x in calls if x]
        calls = [item for sublist in calls for item in sublist]
        self._expression_calls = list(set(calls))

        internal_calls = [x.internal_calls for x in self.nodes]
        internal_calls = [x for x in internal_calls if x]
        internal_calls = [item for sublist in internal_calls for item in sublist]
        self._internal_calls = list(set(internal_calls))

        self._solidity_calls = [
            ir for ir in internal_calls if isinstance(ir.function, SolidityFunction)
        ]

        low_level_calls = [x.low_level_calls for x in self.nodes]
        low_level_calls = [x for x in low_level_calls if x]
        low_level_calls = [item for sublist in low_level_calls for item in sublist]
        self._low_level_calls = list(set(low_level_calls))

        high_level_calls = [x.high_level_calls for x in self.nodes]
        high_level_calls = [x for x in high_level_calls if x]
        high_level_calls = [item for sublist in high_level_calls for item in sublist]
        self._high_level_calls = list(set(high_level_calls))

        library_calls = [x.library_calls for x in self.nodes]
        library_calls = [x for x in library_calls if x]
        library_calls = [item for sublist in library_calls for item in sublist]
        self._library_calls = list(set(library_calls))

        external_calls_as_expressions = [x.external_calls_as_expressions for x in self.nodes]
        external_calls_as_expressions = [x for x in external_calls_as_expressions if x]
        external_calls_as_expressions = [
            item for sublist in external_calls_as_expressions for item in sublist
        ]
        self._external_calls_as_expressions = list(set(external_calls_as_expressions))

    # endregion
    ###################################################################################
    ###################################################################################
    # region Nodes
    ###################################################################################
    ###################################################################################

    def new_node(
        self, node_type: "NodeType", src: str | dict, scope: Union[Scope, "Function"]
    ) -> "Node":
        from slither.core.cfg.node import Node

        node = Node(node_type, self._counter_nodes, scope, self.file_scope)
        node.set_offset(src, self.compilation_unit)
        self._counter_nodes += 1
        node.set_function(self)
        self._nodes.append(node)

        return node

    # endregion
    ###################################################################################
    ###################################################################################
    # region SlithIr and SSA
    ###################################################################################
    ###################################################################################

    def _get_last_ssa_variable_instances(
        self, target_state: bool, target_local: bool
    ) -> dict[str, set["SlithIRVariable"]]:
        from slither.slithir.variables import ReferenceVariable
        from slither.slithir.operations import OperationWithLValue
        from slither.core.cfg.node import NodeType

        if not self.is_implemented:
            return {}

        if self._entry_point is None:
            return {}
        # node, values
        to_explore: list[tuple[Node, dict]] = [(self._entry_point, {})]
        # node -> values
        explored: dict = {}
        # name -> instances
        ret: dict = {}

        while to_explore:
            node, values = to_explore[0]
            to_explore = to_explore[1::]

            if node.type != NodeType.ENTRYPOINT:
                for ir_ssa in node.irs_ssa:
                    if isinstance(ir_ssa, OperationWithLValue):
                        lvalue = ir_ssa.lvalue
                        if isinstance(lvalue, ReferenceVariable):
                            lvalue = lvalue.points_to_origin
                        if isinstance(lvalue, StateVariable) and target_state:
                            values[lvalue.canonical_name] = {lvalue}
                        if isinstance(lvalue, LocalVariable) and target_local:
                            values[lvalue.canonical_name] = {lvalue}

            # Check for fixpoint
            if node in explored:
                if values == explored[node]:
                    continue
                for k, instances in values.items():
                    if k not in explored[node]:
                        explored[node][k] = set()
                    explored[node][k] |= instances
                values = explored[node]
            else:
                explored[node] = values

            # Return condition
            if node.will_return:
                for name, instances in values.items():
                    if name not in ret:
                        ret[name] = set()
                    ret[name] |= instances

            for son in node.sons:
                to_explore.append((son, dict(values)))

        return ret

    def get_last_ssa_state_variables_instances(
        self,
    ) -> dict[str, set["SlithIRVariable"]]:
        return self._get_last_ssa_variable_instances(target_state=True, target_local=False)

    def get_last_ssa_local_variables_instances(
        self,
    ) -> dict[str, set["SlithIRVariable"]]:
        return self._get_last_ssa_variable_instances(target_state=False, target_local=True)

    @staticmethod
    def _unchange_phi(ir: "Operation") -> bool:
        from slither.slithir.operations import Phi, PhiCallback

        if not isinstance(ir, (Phi, PhiCallback)) or len(ir.rvalues) > 1:
            return False
        if not ir.rvalues:
            return True
        return ir.rvalues[0] == ir.lvalue

    def _fix_phi_entry(
        self,
        node: "Node",
        last_state_variables_instances: dict[str, list["StateVariable"]],
        initial_state_variables_instances: dict[str, "StateVariable"],
    ) -> None:
        from slither.slithir.variables import Constant, StateIRVariable, LocalIRVariable

        for ir in node.irs_ssa:
            if isinstance(ir.lvalue, StateIRVariable):
                additional = [initial_state_variables_instances[ir.lvalue.canonical_name]]
                additional += last_state_variables_instances[ir.lvalue.canonical_name]
                ir.rvalues = list(set(additional + ir.rvalues))
            # function parameter that are storage pointer
            else:
                # find index of the parameter
                idx = self.parameters.index(ir.lvalue.non_ssa_version)
                # find non ssa version of that index
                additional = [n.ir.arguments[idx] for n in self.reachable_from_nodes]
                additional = unroll(additional)
                additional = [a for a in additional if not isinstance(a, Constant)]
                ir.rvalues = list(set(additional + ir.rvalues))

                if isinstance(ir.lvalue, LocalIRVariable) and ir.lvalue.is_storage:
                    # Update the refers_to to point to the phi rvalues
                    # This basically means that the local variable is a storage that point to any
                    # state variable that the storage pointer alias analysis found
                    ir.lvalue.refers_to = [
                        rvalue for rvalue in ir.rvalues if isinstance(rvalue, StateIRVariable)
                    ]

    def fix_phi(
        self,
        last_state_variables_instances: dict[str, list["StateVariable"]],
        initial_state_variables_instances: dict[str, "StateVariable"],
    ) -> None:
        from slither.slithir.operations import InternalCall, PhiCallback, Phi
        from slither.slithir.variables import StateIRVariable, LocalIRVariable

        for node in self.nodes:
            if node == self.entry_point:
                self._fix_phi_entry(
                    node, last_state_variables_instances, initial_state_variables_instances
                )
            for ir in node.irs_ssa:
                if isinstance(ir, PhiCallback):
                    callee_ir = ir.callee_ir
                    if isinstance(callee_ir, InternalCall):
                        last_ssa = callee_ir.function.get_last_ssa_state_variables_instances()
                        if ir.lvalue.canonical_name in last_ssa:
                            ir.rvalues = list(last_ssa[ir.lvalue.canonical_name])
                        else:
                            ir.rvalues = [ir.lvalue]
                    else:
                        additional = last_state_variables_instances[ir.lvalue.canonical_name]
                        ir.rvalues = list(set(additional + ir.rvalues))

                # Propage storage ref information if it does not exist
                # This can happen if the refers_to variable was discovered through the phi operator on function parameter
                # aka you have storage pointer as function parameter
                # instead of having a storage pointer for which the aliases belong to the function body
                if (
                    isinstance(ir, Phi)
                    and isinstance(ir.lvalue, LocalIRVariable)
                    and ir.lvalue.is_storage
                    and not ir.lvalue.refers_to
                ):
                    refers_to = []
                    for candidate in ir.rvalues:
                        if isinstance(candidate, StateIRVariable):
                            refers_to.append(candidate)
                        if isinstance(candidate, LocalIRVariable) and candidate.is_storage:
                            refers_to += candidate.refers_to

                    ir.lvalue.refers_to = refers_to

            node.irs_ssa = [ir for ir in node.irs_ssa if not self._unchange_phi(ir)]

    def generate_slithir_and_analyze(self) -> None:
        for node in self.nodes:
            node.slithir_generation()

        self._analyze_read_write()
        self._analyze_calls()

    @abstractmethod
    def generate_slithir_ssa(self, all_ssa_state_variables_instances):
        pass

    def update_read_write_using_ssa(self) -> None:
        for node in self.nodes:
            node.update_read_write_using_ssa()
        self._analyze_read_write()

    ###################################################################################
    ###################################################################################
    # region Built in definitions
    ###################################################################################
    ###################################################################################

    def __str__(self) -> str:
        return self.name

    # endregion



from slither.core.cfg.node import NodeType, Node
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.core.declarations import Contract
from slither.utils.output import Output
from slither.slithir.operations import (
    HighLevelCall,
    LibraryCall,
    LowLevelCall,
    Send,
    Transfer,
    InternalCall,
)

Result = list[tuple[Node, list[str]]]


def detect_call_in_loop(contract: Contract) -> Result:
    ret: Result = []
    for f in contract.functions_entry_points:
        if f.is_implemented:
            call_in_loop(f.entry_point, 0, [], [], ret)

    return ret


def call_in_loop(
    node: Node | None,
    in_loop_counter: int,
    visited: list[Node],
    calls_stack: list[str],
    ret: Result,
) -> None:
    if node is None:
        return
    if node in visited:
        return
    # shared visited
    visited.append(node)

    if node.type == NodeType.STARTLOOP:
        in_loop_counter += 1
    elif node.type == NodeType.ENDLOOP:
        in_loop_counter -= 1

    for ir in node.irs:
        if isinstance(ir, (LowLevelCall, HighLevelCall, Send, Transfer)) and in_loop_counter > 0:
            if isinstance(ir, LibraryCall):
                continue
            ret.append((ir.node, calls_stack.copy()))
        if isinstance(ir, (InternalCall)):
            assert ir.function
            calls_stack.append(node.function.canonical_name)
            call_in_loop(ir.function.entry_point, in_loop_counter, visited, calls_stack, ret)
            calls_stack.pop()

    for son in node.sons:
        call_in_loop(son, in_loop_counter, visited, calls_stack, ret)


class MultipleCallsInLoop(AbstractDetector):
    ARGUMENT = "calls-loop"
    HELP = "Multiple calls in a loop"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation/#calls-inside-a-loop"

    WIKI_TITLE = "Calls inside a loop"
    WIKI_DESCRIPTION = "Calls inside a loop might lead to a denial-of-service attack."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract CallsInLoop{

    address[] destinations;

    constructor(address[] newDestinations) public{
        destinations = newDestinations;
    }

    function bad() external{
        for (uint i=0; i < destinations.length; i++){
            destinations[i].transfer(i);
        }
    }

}
```
If one of the destinations has a fallback function that reverts, `bad` will always revert."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Favor [pull over push](https://github.com/ethereum/wiki/wiki/Safety#favor-pull-over-push-for-external-calls) strategy for external calls."

    def _detect(self) -> list[Output]:
        """"""
        results: list[Output] = []
        for c in self.compilation_unit.contracts_derived:
            values = detect_call_in_loop(c)
            for node, calls_stack in values:
                func = node.function

                info: DETECTOR_INFO = [func, " has external calls inside a loop: ", node, "\n"]

                if len(calls_stack) > 0:
                    info.append("\tCalls stack containing the loop:\n")
                    for call in calls_stack:
                        info.extend(["\t\t", call, "\n"])

                res = self.generate_result(info)
                results.append(res)

        return results



from typing import TYPE_CHECKING

from slither.core.source_mapping.source_mapping import SourceMapping

if TYPE_CHECKING:
    from slither.core.variables.structure_variable import StructureVariable
    from slither.core.compilation_unit import SlitherCompilationUnit


class Structure(SourceMapping):
    def __init__(self, compilation_unit: "SlitherCompilationUnit") -> None:
        super().__init__()
        self._name: str | None = None
        self._canonical_name: str | None = None
        self._elems: dict[str, StructureVariable] = {}
        # Name of the elements in the order of declaration
        self._elems_ordered: list[str] = []
        self.compilation_unit = compilation_unit

    @property
    def canonical_name(self) -> str:
        assert self._canonical_name
        return self._canonical_name

    @canonical_name.setter
    def canonical_name(self, name: str) -> None:
        self._canonical_name = name

    @property
    def name(self) -> str:
        assert self._name
        return self._name

    @name.setter
    def name(self, new_name: str) -> None:
        self._name = new_name

    @property
    def elems(self) -> dict[str, "StructureVariable"]:
        return self._elems

    def add_elem_in_order(self, s: str) -> None:
        self._elems_ordered.append(s)

    @property
    def elems_ordered(self) -> list["StructureVariable"]:
        ret = []
        for e in self._elems_ordered:
            ret.append(self._elems[e])
        return ret

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other) -> bool:
        if not isinstance(other, Structure):
            return False
        if len(self.elems) != len(other.elems):
            return False
        for idx, elem in enumerate(self.elems_ordered):
            other_elem = other.elems_ordered[idx]
            if str(other_elem.type) != str(elem.type) or other_elem.name != elem.name:
                return False
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)



import json
import logging

# https://docs.python.org/3/library/zipfile.html#zipfile-objects
import zipfile
from collections import namedtuple

ZIP_TYPES_ACCEPTED = {
    "lzma": zipfile.ZIP_LZMA,
    "stored": zipfile.ZIP_STORED,
    "deflated": zipfile.ZIP_DEFLATED,
    "bzip2": zipfile.ZIP_BZIP2,
}

Export = namedtuple("Export", ["filename", "content"])

logger = logging.getLogger("Slither-flat")


def save_to_zip(files: list[Export], zip_filename: str, zip_type: str = "lzma"):
    """
    Save projects to a zip
    """
    logger.info(f"Export {zip_filename}")
    with zipfile.ZipFile(
        zip_filename,
        "w",
        compression=ZIP_TYPES_ACCEPTED.get(zip_type, zipfile.ZIP_LZMA),
    ) as file_desc:
        for f in files:
            file_desc.writestr(str(f.filename), f.content)


def save_to_disk(files: list[Export]):
    """
    Save projects to a zip
    """
    for file in files:
        with open(file.filename, "w", encoding="utf8") as f:
            logger.info(f"Export {file.filename}")
            f.write(file.content)


def export_as_json(files: list[Export], filename: str):
    """
    Save projects to a zip
    """

    files_as_dict = {str(f.filename): f.content for f in files}

    if filename == "-":
        print(json.dumps(files_as_dict))
    else:
        logger.info(f"Export {filename}")
        with open(filename, "w", encoding="utf8") as f:
            json.dump(files_as_dict, f)



"""
Detect mistakenly un-indexed ERC20 event parameters
"""

from typing import Any
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations.contract import Contract
from slither.core.declarations.event import Event
from slither.core.variables.event_variable import EventVariable
from slither.utils.output import Output


class UnindexedERC20EventParameters(AbstractDetector):
    """
    Un-indexed ERC20 event parameters
    """

    ARGUMENT = "erc20-indexed"
    HELP = "Un-indexed ERC20 event parameters"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#unindexed-erc20-event-parameters"

    WIKI_TITLE = "Unindexed ERC20 event parameters"
    WIKI_DESCRIPTION = "Detects whether events defined by the `ERC20` specification that should have some parameters as `indexed` are missing the `indexed` keyword."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract ERC20Bad {
    // ...
    event Transfer(address from, address to, uint value);
    event Approval(address owner, address spender, uint value);

    // ...
}
```
`Transfer` and `Approval` events should have the 'indexed' keyword on their two first parameters, as defined by the `ERC20` specification.
Failure to include these keywords will exclude the parameter data in the transaction/block's bloom filter, so external tooling searching for these parameters may overlook them and fail to index logs from this token contract."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Add the `indexed` keyword to event parameters that should include it, according to the `ERC20` specification."

    STANDARD_JSON = False

    @staticmethod
    def detect_erc20_unindexed_event_params(
        contract: Contract,
    ) -> list[tuple[Event, EventVariable] | Any]:
        """
        Detect un-indexed ERC20 event parameters in a given contract.
        :param contract: The contract to check ERC20 events for un-indexed parameters in.
        :return: A list of tuple(event, parameter) of parameters which should be indexed.
        """
        # Create our result array
        results = []

        # If this contract isn't an ERC20 token, we return our empty results.
        if not contract.is_erc20():
            return results

        # Loop through all events to look for poor form.
        for event in contract.events_declared:
            # If this is transfer/approval events, expect the first two parameters to be indexed.
            if event.full_name in [
                "Transfer(address,address,uint256)",
                "Approval(address,address,uint256)",
            ]:
                if not event.elems[0].indexed:
                    results.append((event, event.elems[0]))
                if not event.elems[1].indexed:
                    results.append((event, event.elems[1]))

        # Return the results.
        return results

    def _detect(self) -> list[Output]:
        """
        Detect un-indexed ERC20 event parameters in all contracts.
        """
        results = []
        for c in self.contracts:
            unindexed_params = self.detect_erc20_unindexed_event_params(c)
            if unindexed_params:
                # Add each problematic event definition to our result list
                for event, parameter in unindexed_params:
                    info = [
                        "ERC20 event ",
                        event,
                        f"does not index parameter {parameter}\n",
                    ]

                    # Add the events to the JSON (note: we do not add the params/vars as they have no source mapping).
                    res = self.generate_result(info)

                    res.add(event, {"parameter_name": parameter.name})
                    results.append(res)

        return results



from slither.core.declarations import Function
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.code_complexity import compute_cyclomatic_complexity
from slither.utils.output import Output


def _check_for_high_cc(high_cc_functions: list[tuple[Function, int]], f: Function) -> None:
    cc = compute_cyclomatic_complexity(f)
    if cc > 11:
        high_cc_functions.append((f, cc))


class CyclomaticComplexity(AbstractDetector):
    """
    Detects functions with high (> 11) cyclomatic complexity.
    """

    ARGUMENT = "cyclomatic-complexity"
    HELP = "Detects functions with high (> 11) cyclomatic complexity"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#cyclomatic-complexity"

    WIKI_TITLE = "Cyclomatic complexity"
    WIKI_DESCRIPTION = "Detects functions with high (> 11) cyclomatic complexity."
    WIKI_EXPLOIT_SCENARIO = ""
    WIKI_RECOMMENDATION = (
        "Reduce cyclomatic complexity by splitting the function into several smaller subroutines."
    )

    def _detect(self) -> list[Output]:
        results = []
        high_cc_functions: list[tuple[Function, int]] = []

        f: Function
        for c in self.compilation_unit.contracts:
            for f in c.functions_declared:
                _check_for_high_cc(high_cc_functions, f)

        for f in self.compilation_unit.functions_top_level:
            _check_for_high_cc(high_cc_functions, f)

        for f, cc in high_cc_functions:
            info: DETECTOR_INFO = [f, f" has a high cyclomatic complexity ({cc}).\n"]
            res = self.generate_result(info)
            results.append(res)
        return results



from slither.core.variables.variable import Variable
from slither.slithir.operations import Operation


class Nop(Operation):
    @property
    def read(self) -> list[Variable]:
        return []

    @property
    def used(self):
        return []

    def __str__(self):
        return "NOP"



from typing import TYPE_CHECKING
from slither.core.declarations.contract_level import ContractLevel


from slither.core.declarations.custom_error import CustomError

if TYPE_CHECKING:
    from slither.core.declarations import Contract


class CustomErrorContract(CustomError, ContractLevel):
    def is_declared_by(self, contract: "Contract") -> bool:
        """
        Check if the element is declared by the contract
        :param contract:
        :return:
        """
        return self.contract == contract

    @property
    def canonical_name(self) -> str:
        return self.contract.name + "." + self.full_name



from .constant import Constant
from .reference import ReferenceVariable
from .reference_ssa import ReferenceVariableSSA
from .temporary import TemporaryVariable
from .temporary_ssa import TemporaryVariableSSA
from .tuple import TupleVariable
from .tuple_ssa import TupleVariableSSA
from .local_variable import LocalIRVariable
from .state_variable import StateIRVariable



import logging
from enum import Enum

from slither.core.expressions.expression import Expression
from slither.core.exceptions import SlitherCoreError


logger = logging.getLogger("BinaryOperation")


class BinaryOperationType(Enum):
    POWER = 0  # **
    MULTIPLICATION = 1  # *
    DIVISION = 2  # /
    MODULO = 3  # %
    ADDITION = 4  # +
    SUBTRACTION = 5  # -
    LEFT_SHIFT = 6  # <<
    RIGHT_SHIFT = 7  # >>>
    AND = 8  # &
    CARET = 9  # ^
    OR = 10  # |
    LESS = 11  # <
    GREATER = 12  # >
    LESS_EQUAL = 13  # <=
    GREATER_EQUAL = 14  # >=
    EQUAL = 15  # ==
    NOT_EQUAL = 16  # !=
    ANDAND = 17  # &&
    OROR = 18  # ||

    # YUL specific operators
    # TODO: investigate if we can remove these
    # Find the types earlier on, and do the conversion
    DIVISION_SIGNED = 19
    MODULO_SIGNED = 20
    LESS_SIGNED = 21
    GREATER_SIGNED = 22
    RIGHT_SHIFT_ARITHMETIC = 23

    @staticmethod
    def get_type(
        operation_type: "str",
    ) -> "BinaryOperationType":
        if operation_type == "**":
            return BinaryOperationType.POWER
        if operation_type == "*":
            return BinaryOperationType.MULTIPLICATION
        if operation_type == "/":
            return BinaryOperationType.DIVISION
        if operation_type == "%":
            return BinaryOperationType.MODULO
        if operation_type == "+":
            return BinaryOperationType.ADDITION
        if operation_type == "-":
            return BinaryOperationType.SUBTRACTION
        if operation_type == "<<":
            return BinaryOperationType.LEFT_SHIFT
        if operation_type == ">>":
            return BinaryOperationType.RIGHT_SHIFT
        if operation_type == "&":
            return BinaryOperationType.AND
        if operation_type == "^":
            return BinaryOperationType.CARET
        if operation_type == "|":
            return BinaryOperationType.OR
        if operation_type == "<":
            return BinaryOperationType.LESS
        if operation_type == ">":
            return BinaryOperationType.GREATER
        if operation_type == "<=":
            return BinaryOperationType.LESS_EQUAL
        if operation_type == ">=":
            return BinaryOperationType.GREATER_EQUAL
        if operation_type == "==":
            return BinaryOperationType.EQUAL
        if operation_type == "!=":
            return BinaryOperationType.NOT_EQUAL
        if operation_type == "&&":
            return BinaryOperationType.ANDAND
        if operation_type == "||":
            return BinaryOperationType.OROR
        if operation_type == "/'":
            return BinaryOperationType.DIVISION_SIGNED
        if operation_type == "%'":
            return BinaryOperationType.MODULO_SIGNED
        if operation_type == "<'":
            return BinaryOperationType.LESS_SIGNED
        if operation_type == ">'":
            return BinaryOperationType.GREATER_SIGNED
        if operation_type == ">>'":
            return BinaryOperationType.RIGHT_SHIFT_ARITHMETIC

        raise SlitherCoreError(f"get_type: Unknown operation type {operation_type})")

    def __str__(self) -> str:
        if self == BinaryOperationType.POWER:
            return "**"
        if self == BinaryOperationType.MULTIPLICATION:
            return "*"
        if self == BinaryOperationType.DIVISION:
            return "/"
        if self == BinaryOperationType.MODULO:
            return "%"
        if self == BinaryOperationType.ADDITION:
            return "+"
        if self == BinaryOperationType.SUBTRACTION:
            return "-"
        if self == BinaryOperationType.LEFT_SHIFT:
            return "<<"
        if self == BinaryOperationType.RIGHT_SHIFT:
            return ">>"
        if self == BinaryOperationType.AND:
            return "&"
        if self == BinaryOperationType.CARET:
            return "^"
        if self == BinaryOperationType.OR:
            return "|"
        if self == BinaryOperationType.LESS:
            return "<"
        if self == BinaryOperationType.GREATER:
            return ">"
        if self == BinaryOperationType.LESS_EQUAL:
            return "<="
        if self == BinaryOperationType.GREATER_EQUAL:
            return ">="
        if self == BinaryOperationType.EQUAL:
            return "=="
        if self == BinaryOperationType.NOT_EQUAL:
            return "!="
        if self == BinaryOperationType.ANDAND:
            return "&&"
        if self == BinaryOperationType.OROR:
            return "||"
        if self == BinaryOperationType.DIVISION_SIGNED:
            return "/'"
        if self == BinaryOperationType.MODULO_SIGNED:
            return "%'"
        if self == BinaryOperationType.LESS_SIGNED:
            return "<'"
        if self == BinaryOperationType.GREATER_SIGNED:
            return ">'"
        if self == BinaryOperationType.RIGHT_SHIFT_ARITHMETIC:
            return ">>'"
        raise SlitherCoreError(f"str: Unknown operation type {self})")


class BinaryOperation(Expression):
    def __init__(
        self,
        left_expression: Expression,
        right_expression: Expression,
        expression_type: BinaryOperationType,
    ) -> None:
        assert isinstance(left_expression, Expression)
        assert isinstance(right_expression, Expression)
        super().__init__()
        self._expressions = [left_expression, right_expression]
        self._type: BinaryOperationType = expression_type

    @property
    def expressions(self) -> list[Expression]:
        return self._expressions

    @property
    def expression_left(self) -> Expression:
        return self._expressions[0]

    @property
    def expression_right(self) -> Expression:
        return self._expressions[1]

    @property
    def type(self) -> BinaryOperationType:
        return self._type

    def __str__(self) -> str:
        return str(self.expression_left) + " " + str(self.type) + " " + str(self.expression_right)



from typing import TYPE_CHECKING

from slither.core.cfg.node import NodeType

if TYPE_CHECKING:
    from slither.core.cfg.node import Node


def intersection_predecessor(node: "Node") -> set["Node"]:
    if not node.fathers:
        return set()

    ret = node.fathers[0].dominators
    for pred in node.fathers[1:]:
        ret = ret.intersection(pred.dominators)
    if not any(father.is_reachable for father in node.fathers):
        return set()

    ret = set()
    for pred in node.fathers:
        ret = ret.union(pred.dominators)

    for pred in node.fathers:
        if pred.is_reachable:
            ret = ret.intersection(pred.dominators)
    return ret


def _compute_dominators(nodes: list["Node"]) -> None:
    changed = True

    while changed:
        changed = False

        for node in nodes:
            new_set = intersection_predecessor(node).union({node})
            if new_set != node.dominators:
                node.dominators = new_set
                changed = True


def _compute_immediate_dominators(nodes: list["Node"]) -> None:
    for node in nodes:
        idom_candidates = set(node.dominators)
        idom_candidates.remove(node)

        if len(idom_candidates) == 1:
            idom = idom_candidates.pop()
            node.immediate_dominator = idom
            idom.dominator_successors.add(node)
            continue

        # all_dominators contain all the dominators of all the node's dominators
        # But self inclusion is removed
        # The idom is then the only node that in idom_candidate that is not in all_dominators
        all_dominators = set()
        for d in idom_candidates:
            # optimization: if a node is already in all_dominators, then
            # its dominators are already in too
            if d in all_dominators:
                continue
            all_dominators |= d.dominators - {d}

        idom_candidates = all_dominators.symmetric_difference(idom_candidates)
        assert len(idom_candidates) <= 1
        if idom_candidates:
            idom = idom_candidates.pop()
            node.immediate_dominator = idom
            idom.dominator_successors.add(node)


def compute_dominators(nodes: list["Node"]) -> None:
    """
    Naive implementation of Cooper, Harvey, Kennedy algo
    See 'A Simple,Fast Dominance Algorithm'

    Compute strict domniators
    """

    for n in nodes:
        n.dominators = set(nodes)

    _compute_dominators(nodes)

    _compute_immediate_dominators(nodes)


def compute_dominance_frontier(nodes: list["Node"]) -> None:
    """
    Naive implementation of Cooper, Harvey, Kennedy algo
    See 'A Simple,Fast Dominance Algorithm'

    Compute dominance frontier
    """
    for node in nodes:
        if len(node.fathers) >= 2:
            for father in node.fathers:
                if not father.is_reachable:
                    continue
                runner = father
                # Corner case: if there is a if without else
                # we need to add update the conditional node
                if (
                    runner == node.immediate_dominator
                    and runner.type == NodeType.IF
                    and node.type == NodeType.ENDIF
                ):
                    runner.dominance_frontier = runner.dominance_frontier.union({node})
                while runner != node.immediate_dominator:
                    runner.dominance_frontier = runner.dominance_frontier.union({node})
                    assert runner.immediate_dominator
                    runner = runner.immediate_dominator



import logging

from slither.core.declarations import Contract
from slither.slithir.operations import EventCall
from slither.utils import output
from slither.utils.erc import ERC, ERC_EVENT
from slither.utils.type import (
    export_nested_types_from_variable,
    export_return_type_from_variable,
)

logger = logging.getLogger("Slither-conformance")


def _check_signature(erc_function: ERC, contract: Contract, ret: dict) -> None:
    name = erc_function.name
    parameters = erc_function.parameters
    return_type = erc_function.return_type
    view = erc_function.view
    required = erc_function.required
    events = erc_function.events

    sig = f"{name}({','.join(parameters)})"
    function = contract.get_function_from_signature(sig)

    if not function:
        # The check on state variable is needed until we have a better API to handle state variable getters
        state_variable_as_function = contract.get_state_variable_from_name(name)

        if not state_variable_as_function or state_variable_as_function.visibility not in [
            "public",
            "external",
        ]:
            txt = f"[ ] {sig} is missing {'' if required else '(optional)'}"
            logger.info(txt)
            missing_func = output.Output(
                txt, additional_fields={"function": sig, "required": required}
            )
            missing_func.add(contract)
            ret["missing_function"].append(missing_func.data)
            return

        types = [str(x) for x in export_nested_types_from_variable(state_variable_as_function)]

        if types != parameters:
            txt = f"[ ] {sig} is missing {'' if required else '(optional)'}"
            logger.info(txt)
            missing_func = output.Output(
                txt, additional_fields={"function": sig, "required": required}
            )
            missing_func.add(contract)
            ret["missing_function"].append(missing_func.data)
            return

        function_return_type = export_return_type_from_variable(state_variable_as_function)
        function = state_variable_as_function

        function_view = True
    else:
        function_return_type = function.return_type
        function_view = function.view

    txt = f"[✓] {sig} is present"
    logger.info(txt)

    if function_return_type:
        function_return_type = ",".join([str(x) for x in function_return_type])
        if function_return_type == return_type:
            txt = f"\t[✓] {sig} -> ({function_return_type}) (correct return type)"
            logger.info(txt)
        else:
            txt = f"\t[ ] {sig} -> ({function_return_type}) should return {return_type}"
            logger.info(txt)

            incorrect_return = output.Output(
                txt,
                additional_fields={
                    "expected_return_type": return_type,
                    "actual_return_type": function_return_type,
                },
            )
            incorrect_return.add(function)
            ret["incorrect_return_type"].append(incorrect_return.data)

    elif not return_type:
        txt = f"\t[✓] {sig} -> () (correct return type)"
        logger.info(txt)
    else:
        txt = f"\t[ ] {sig} -> () should return {return_type}"
        logger.info(txt)

        incorrect_return = output.Output(
            txt,
            additional_fields={
                "expected_return_type": return_type,
                "actual_return_type": function_return_type,
            },
        )
        incorrect_return.add(function)
        ret["incorrect_return_type"].append(incorrect_return.data)

    if view:
        if function_view:
            txt = f"\t[✓] {sig} is view"
            logger.info(txt)
        else:
            txt = f"\t[ ] {sig} should be view"
            logger.info(txt)

            should_be_view = output.Output(txt)
            should_be_view.add(function)
            ret["should_be_view"].append(should_be_view.data)

    if events:
        for event in events:
            event_sig = f"{event.name}({','.join(event.parameters)})"

            if not function:
                txt = f"\t[ ] Must emit be view {event_sig}"
                logger.info(txt)

                missing_event_emmited = output.Output(
                    txt, additional_fields={"missing_event": event_sig}
                )
                missing_event_emmited.add(function)
                ret["missing_event_emmited"].append(missing_event_emmited.data)

            else:
                event_found = False
                for ir in function.all_slithir_operations():
                    if isinstance(ir, EventCall):
                        if ir.name == event.name:
                            if event.parameters == [str(a.type) for a in ir.arguments]:
                                event_found = True
                                break
                if event_found:
                    txt = f"\t[✓] {event_sig} is emitted"
                    logger.info(txt)
                else:
                    txt = f"\t[ ] Must emit be view {event_sig}"
                    logger.info(txt)

                    missing_event_emmited = output.Output(
                        txt, additional_fields={"missing_event": event_sig}
                    )
                    missing_event_emmited.add(function)
                    ret["missing_event_emmited"].append(missing_event_emmited.data)


def _check_events(erc_event: ERC_EVENT, contract: Contract, ret: dict[str, list]) -> None:
    name = erc_event.name
    parameters = erc_event.parameters
    indexes = erc_event.indexes

    sig = f"{name}({','.join(parameters)})"
    event = contract.get_event_from_signature(sig)

    if not event:
        txt = f"[ ] {sig} is missing"
        logger.info(txt)

        missing_event = output.Output(txt, additional_fields={"event": sig})
        missing_event.add(contract)
        ret["missing_event"].append(missing_event.data)
        return

    txt = f"[✓] {sig} is present"
    logger.info(txt)

    for i, index in enumerate(indexes):
        if index:
            if event.elems[i].indexed:
                txt = f"\t[✓] parameter {i} is indexed"
                logger.info(txt)
            else:
                txt = f"\t[ ] parameter {i} should be indexed"
                logger.info(txt)

                missing_event_index = output.Output(txt, additional_fields={"missing_index": i})
                missing_event_index.add_event(event)
                ret["missing_event_index"].append(missing_event_index.data)


def generic_erc_checks(
    contract: Contract,
    erc_functions: list[ERC],
    erc_events: list[ERC_EVENT],
    ret: dict[str, list],
    explored: set[Contract] | None = None,
) -> None:
    if explored is None:
        explored = set()

    explored.add(contract)

    logger.info(f"# Check {contract.name}\n")

    logger.info("## Check functions")
    for erc_function in erc_functions:
        _check_signature(erc_function, contract, ret)
    if erc_events:
        logger.info("\n## Check events")
        for erc_event in erc_events:
            _check_events(erc_event, contract, ret)

    logger.info("\n")

    for derived_contract in contract.derived_contracts:
        generic_erc_checks(derived_contract, erc_functions, erc_events, ret, explored)



"""
Function module
"""

from typing import TYPE_CHECKING

from slither.core.declarations.contract_level import ContractLevel
from slither.core.declarations import Function
from slither.utils.code_complexity import compute_cyclomatic_complexity


if TYPE_CHECKING:
    from slither.core.declarations import Contract
    from slither.core.scope.scope import FileScope
    from slither.slithir.variables.state_variable import StateIRVariable
    from slither.core.compilation_unit import SlitherCompilationUnit


class FunctionContract(Function, ContractLevel):
    def __init__(self, compilation_unit: "SlitherCompilationUnit") -> None:
        super().__init__(compilation_unit)
        self._contract_declarer: Contract | None = None

    def set_contract_declarer(self, contract: "Contract") -> None:
        self._contract_declarer = contract

    @property
    def contract_declarer(self) -> "Contract":
        """
        Return the contract where this function was declared. Only functions have both a contract, and contract_declarer
        This is because we need to have separate representation of the function depending of the contract's context
        For example a function calling super.f() will generate different IR depending on the current contract's inheritance

        Returns:
            The contract where this function was declared
        """

        assert self._contract_declarer
        return self._contract_declarer

    @property
    def canonical_name(self) -> str:
        """
        str: contract.func_name(type1,type2)
        Return the function signature without the return values
        """
        if self._canonical_name is None:
            name, parameters, _ = self.signature
            self._canonical_name = (
                ".".join([self.contract_declarer.name] + self._internal_scope + [name])
                + "("
                + ",".join(parameters)
                + ")"
            )
        return self._canonical_name

    def is_declared_by(self, contract: "Contract") -> bool:
        """
        Check if the element is declared by the contract
        :param contract:
        :return:
        """
        return self.contract_declarer == contract

    @property
    def file_scope(self) -> "FileScope":
        # This is the contract declarer's file scope because inherited functions have access
        # to the file scope which their declared in. This scope may contain references not
        # available in the child contract's scope. See inherited_function_scope.sol for an example.
        return self.contract_declarer.file_scope

    # endregion
    ###################################################################################
    ###################################################################################
    # region Functions
    ###################################################################################
    ###################################################################################

    @property
    def functions_shadowed(self) -> list["Function"]:
        """
            Return the list of functions shadowed
        Returns:
            list(core.Function)

        """
        candidates = [c.functions_declared for c in self.contract.inheritance]
        candidates = [candidate for sublist in candidates for candidate in sublist]
        return [f for f in candidates if f.full_name == self.full_name]

    # endregion
    ###################################################################################
    ###################################################################################
    # region Summary information
    ###################################################################################
    ###################################################################################

    def get_summary(
        self,
    ) -> tuple[str, str, str, list[str], list[str], list[str], list[str], list[str], int]:
        """
            Return the function summary
        Returns:
            (str, str, str, list(str), list(str), listr(str), list(str), list(str);
            contract_name, name, visibility, modifiers, vars read, vars written, internal_calls, external_calls_as_expressions
        """
        return (
            self.contract_declarer.name,
            self.full_name,
            self.visibility,
            [str(x) for x in self.modifiers],
            [str(x) for x in self.state_variables_read + self.solidity_variables_read],
            [str(x) for x in self.state_variables_written],
            [str(x) for x in self.internal_calls],
            [str(x) for x in self.external_calls_as_expressions],
            compute_cyclomatic_complexity(self),
        )

    # endregion
    ###################################################################################
    ###################################################################################
    # region SlithIr and SSA
    ###################################################################################
    ###################################################################################

    def generate_slithir_ssa(
        self, all_ssa_state_variables_instances: dict[str, "StateIRVariable"]
    ) -> None:
        from slither.slithir.utils.ssa import add_ssa_ir, transform_slithir_vars_to_ssa
        from slither.core.dominators.utils import (
            compute_dominance_frontier,
            compute_dominators,
        )

        compute_dominators(self.nodes)
        compute_dominance_frontier(self.nodes)
        transform_slithir_vars_to_ssa(self)
        if not self.contract.is_incorrectly_constructed:
            add_ssa_ir(self, all_ssa_state_variables_instances)



from typing import Any
from slither.core.declarations.solidity_variables import SolidityFunction
from slither.core.solidity_types.elementary_type import ElementaryType
from slither.slithir.operations.call import Call
from slither.slithir.operations.lvalue import OperationWithLValue


class SolidityCall(Call, OperationWithLValue):
    def __init__(
        self,
        function: SolidityFunction,
        nbr_arguments: int,
        result,
        type_call: str | list[ElementaryType],
    ) -> None:
        assert isinstance(function, SolidityFunction)
        super().__init__()
        self._function = function
        self._nbr_arguments = nbr_arguments
        self._type_call = type_call
        self._lvalue = result

    @property
    def read(self) -> list[Any]:
        return self._unroll(self.arguments)

    @property
    def function(self) -> SolidityFunction:
        return self._function

    @property
    def nbr_arguments(self) -> int:
        return self._nbr_arguments

    @property
    def type_call(self) -> str | list[ElementaryType]:
        return self._type_call

    def __str__(self):
        if (
            self.function == SolidityFunction("abi.decode()")
            and len(self.arguments) == 2
            and isinstance(self.arguments[1], list)
        ):
            args = (
                str(self.arguments[0]) + "(" + ",".join([str(a) for a in self.arguments[1]]) + ")"
            )
        else:
            args = ",".join([str(a) for a in self.arguments])

        lvalue = ""
        if self.lvalue:
            if isinstance(self.lvalue.type, (list,)):
                lvalue = f"{self.lvalue}({','.join(str(x) for x in self.lvalue.type)}) = "
            else:
                lvalue = f"{self.lvalue}({self.lvalue.type}) = "
        return lvalue + f"SOLIDITY_CALL {self.function.full_name}({args})"



import json
from pathlib import Path
from subprocess import PIPE, Popen
from slither import Slither

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"
STORAGE_TEST_ROOT = Path(TEST_DATA_DIR, "storage_layout")


def test_storage_layout(solc_binary_path):
    _test_storage_layout(solc_binary_path, "0.8.10", "storage_layout-0.8.10.sol")


def test_custom_storage_layout(solc_binary_path):
    _test_storage_layout(solc_binary_path, "0.8.33", "storage_layout-0.8.33.sol")


def _test_storage_layout(solc_binary_path, solc_version, test_file):
    # the storage layout has not yet changed between solidity versions so we will test with one version of the compiler
    solc_path = solc_binary_path(solc_version)
    test_item = Path(STORAGE_TEST_ROOT, test_file).as_posix()

    sl = Slither(test_item, disallow_partial=True, solc=solc_path)

    with Popen([solc_path, test_item, "--storage-layout"], stdout=PIPE) as process:
        for line in process.stdout:  # parse solc output
            if '{"storage":[{' in line.decode("utf-8"):  # find the storage layout
                layout = iter(json.loads(line)["storage"])
                while True:
                    try:
                        for contract in sl.contracts:
                            curr_var = next(layout)
                            var_name = curr_var["label"]
                            sl_name = contract.variables_as_dict[var_name]
                            slot, offset = contract.compilation_unit.storage_layout_of(
                                contract, sl_name
                            )
                            assert slot == int(curr_var["slot"])
                            assert offset == int(curr_var["offset"])
                    except StopIteration:
                        break
                    except KeyError as e:
                        print(f"not found {e} ")
        process.communicate()
        assert process.returncode == 0



from slither.core.variables.event_variable import EventVariable
from slither.vyper_parsing.type_parsing import parse_type
from slither.vyper_parsing.ast.types import AnnAssign, Call


class EventVariableVyper:
    def __init__(self, variable: EventVariable, variable_data: AnnAssign):
        self._variable = variable
        self._variable.name = variable_data.target.id
        if (
            isinstance(variable_data.annotation, Call)
            and variable_data.annotation.func.id == "indexed"
        ):
            self._variable.indexed = True
        else:
            self._variable.indexed = False
        self._elem_to_parse = variable_data.annotation

    @property
    def underlying_variable(self) -> EventVariable:
        return self._variable

    def analyze(self, contract) -> None:
        self._variable.type = parse_type(self._elem_to_parse, contract)



from typing import TYPE_CHECKING
from slither.core.variables.variable import Variable


if TYPE_CHECKING:
    from slither.core.declarations import Structure


class StructureVariable(Variable):
    def __init__(self) -> None:
        super().__init__()
        self._structure: Structure | None = None

    def set_structure(self, structure: "Structure") -> None:
        self._structure = structure

    @property
    def structure(self) -> "Structure":
        return self._structure



"""
Lines of Code (LOC) printer

Definitions:
cloc: comment lines of code containing only comments
sloc: source lines of code with no whitespace or comments
loc: all lines of code including whitespace and comments
src: source files (excluding tests and dependencies)
dep: dependency files
test: test files
"""

from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.loc import compute_loc_metrics
from slither.utils.output import Output


class LocPrinter(AbstractPrinter):
    ARGUMENT = "loc"
    HELP = """Count the total number lines of code (LOC), source lines of code (SLOC),
            and comment lines of code (CLOC) found in source files (SRC), dependencies (DEP),
            and test files (TEST)."""

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#loc"

    def output(self, _filename: str) -> Output:
        # compute loc metrics
        loc = compute_loc_metrics(self.slither)

        table = loc.to_pretty_table()
        txt = "Lines of Code \n" + str(table)
        self.info(txt)
        res = self.generate_output(txt)
        res.add_pretty_table(table, "Code Lines")
        return res



import math
from enum import Enum
from typing import TYPE_CHECKING

from crytic_compile import CompilationUnit, CryticCompile
from crytic_compile.compiler.compiler import CompilerVersion
from crytic_compile.utils.naming import Filename

from slither.core.context.context import Context
from slither.core.declarations import (
    Contract,
    Pragma,
    Import,
    Function,
    Modifier,
)
from slither.core.declarations.custom_error_top_level import CustomErrorTopLevel
from slither.core.declarations.enum_top_level import EnumTopLevel
from slither.core.declarations.event_top_level import EventTopLevel
from slither.core.declarations.function_top_level import FunctionTopLevel
from slither.core.declarations.structure_top_level import StructureTopLevel
from slither.core.declarations.using_for_top_level import UsingForTopLevel
from slither.core.scope.scope import FileScope
from slither.core.solidity_types.type_alias import TypeAliasTopLevel
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.top_level_variable import TopLevelVariable
from slither.slithir.operations import InternalCall
from slither.slithir.variables import Constant

if TYPE_CHECKING:
    from slither.core.slither_core import SlitherCore


class Language(Enum):
    SOLIDITY = "solidity"
    VYPER = "vyper"

    @staticmethod
    def from_str(label: str):
        if label == "solc":
            return Language.SOLIDITY
        if label == "vyper":
            return Language.VYPER

        raise ValueError(f"Unknown language: {label}")


class SlitherCompilationUnit(Context):
    def __init__(self, core: "SlitherCore", crytic_compilation_unit: CompilationUnit) -> None:
        super().__init__()

        self._core = core
        self._crytic_compile_compilation_unit = crytic_compilation_unit
        self._language = Language.from_str(crytic_compilation_unit.compiler_version.compiler)

        # Top level object
        self.contracts: list[Contract] = []
        self._structures_top_level: list[StructureTopLevel] = []
        self._enums_top_level: list[EnumTopLevel] = []
        self._events_top_level: list[EventTopLevel] = []
        self._variables_top_level: list[TopLevelVariable] = []
        self._functions_top_level: list[FunctionTopLevel] = []
        self._using_for_top_level: list[UsingForTopLevel] = []
        self._pragma_directives: list[Pragma] = []
        self._import_directives: list[Import] = []
        self._custom_errors: list[CustomErrorTopLevel] = []
        self._type_aliases: dict[str, TypeAliasTopLevel] = {}

        self._all_functions: set[Function] = set()
        self._all_modifiers: set[Modifier] = set()

        # Memoize
        self._all_state_variables: set[StateVariable] | None = None

        self._persistent_storage_layouts: dict[str, dict[str, tuple[int, int]]] = {}
        self._transient_storage_layouts: dict[str, dict[str, tuple[int, int]]] = {}

        self._contract_with_missing_inheritance: set[Contract] = set()

        self._source_units: dict[int, str] = {}

        self.counter_slithir_tuple = 0
        self.counter_slithir_temporary = 0
        self.counter_slithir_reference = 0

        self.scopes: dict[Filename, FileScope] = {}

    @property
    def core(self) -> "SlitherCore":
        return self._core

    @property
    def source_units(self) -> dict[int, str]:
        return self._source_units

    # endregion
    ###################################################################################
    ###################################################################################
    # region Compiler
    ###################################################################################
    ###################################################################################
    @property
    def language(self) -> Language:
        return self._language

    @property
    def is_vyper(self) -> bool:
        return self._language == Language.VYPER

    @property
    def is_solidity(self) -> bool:
        return self._language == Language.SOLIDITY

    @property
    def compiler_version(self) -> CompilerVersion:
        return self._crytic_compile_compilation_unit.compiler_version

    @property
    def solc_version(self) -> str:
        # TODO: make version a non optional argument of compiler version in cc
        return self._crytic_compile_compilation_unit.compiler_version.version  # type:ignore

    @property
    def crytic_compile_compilation_unit(self) -> CompilationUnit:
        return self._crytic_compile_compilation_unit

    @property
    def crytic_compile(self) -> CryticCompile:
        return self._crytic_compile_compilation_unit.crytic_compile

    # endregion
    ###################################################################################
    ###################################################################################
    # region Pragma attributes
    ###################################################################################
    ###################################################################################

    @property
    def pragma_directives(self) -> list[Pragma]:
        """list(core.declarations.Pragma): Pragma directives."""
        return self._pragma_directives

    @property
    def import_directives(self) -> list[Import]:
        """list(core.declarations.Import): Import directives"""
        return self._import_directives

    # endregion
    ###################################################################################
    ###################################################################################
    # region Contracts
    ###################################################################################
    ###################################################################################

    @property
    def contracts_derived(self) -> list[Contract]:
        """list(Contract): List of contracts that are derived and not inherited."""
        inheritances = [x.inheritance for x in self.contracts]
        inheritance = [item for sublist in inheritances for item in sublist]
        return [c for c in self.contracts if c not in inheritance]

    def get_contract_from_name(self, contract_name: str | Constant) -> list[Contract]:
        """
            Return a list of contract from a name
        Args:
            contract_name (str): name of the contract
        Returns:
            List[Contract]
        """
        return [c for c in self.contracts if c.name == contract_name]

    # endregion
    ###################################################################################
    ###################################################################################
    # region Functions and modifiers
    ###################################################################################
    ###################################################################################

    @property
    def functions(self) -> list[Function]:
        return list(self._all_functions)

    def add_function(self, func: Function) -> None:
        self._all_functions.add(func)

    @property
    def modifiers(self) -> list[Modifier]:
        return list(self._all_modifiers)

    def add_modifier(self, modif: Modifier) -> None:
        self._all_modifiers.add(modif)

    @property
    def functions_and_modifiers(self) -> list[Function]:
        return self.functions + list(self.modifiers)

    def propagate_function_calls(self) -> None:
        """This info is used to compute the rvalues of Phi operations in `fix_phi` and ultimately
        is responsible for the `read` property of Phi operations which is vital to
        propagating taints inter-procedurally
        """
        for f in self.functions_and_modifiers:
            for node in f.nodes:
                for ir in node.irs_ssa:
                    if isinstance(ir, InternalCall):
                        assert ir.function
                        ir.function.add_reachable_from_node(node, ir)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Variables
    ###################################################################################
    ###################################################################################

    @property
    def state_variables(self) -> list[StateVariable]:
        if self._all_state_variables is None:
            state_variabless = [c.state_variables for c in self.contracts]
            state_variables = [item for sublist in state_variabless for item in sublist]
            self._all_state_variables = set(state_variables)
        return list(self._all_state_variables)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Top level
    ###################################################################################
    ###################################################################################

    @property
    def structures_top_level(self) -> list[StructureTopLevel]:
        return self._structures_top_level

    @property
    def enums_top_level(self) -> list[EnumTopLevel]:
        return self._enums_top_level

    @property
    def events_top_level(self) -> list[EventTopLevel]:
        return self._events_top_level

    @property
    def variables_top_level(self) -> list[TopLevelVariable]:
        return self._variables_top_level

    @property
    def functions_top_level(self) -> list[FunctionTopLevel]:
        return self._functions_top_level

    @property
    def using_for_top_level(self) -> list[UsingForTopLevel]:
        return self._using_for_top_level

    @property
    def custom_errors(self) -> list[CustomErrorTopLevel]:
        return self._custom_errors

    @property
    def type_aliases(self) -> dict[str, TypeAliasTopLevel]:
        return self._type_aliases

    # endregion
    ###################################################################################
    ###################################################################################
    # region Internals
    ###################################################################################
    ###################################################################################

    @property
    def contracts_with_missing_inheritance(self) -> set[Contract]:
        return self._contract_with_missing_inheritance

    # endregion
    ###################################################################################
    ###################################################################################
    # region Scope
    ###################################################################################
    ###################################################################################

    def get_scope(self, filename_str: str) -> FileScope:
        filename = self._crytic_compile_compilation_unit.crytic_compile.filename_lookup(
            filename_str
        )

        if filename not in self.scopes:
            self.scopes[filename] = FileScope(filename)

        return self.scopes[filename]

    # endregion
    ###################################################################################
    ###################################################################################
    # region Storage Layouts
    ###################################################################################
    ###################################################################################

    def compute_storage_layout(self) -> None:
        assert self.is_solidity

        for contract in self.contracts_derived:
            self._compute_storage_layout(
                contract.name,
                contract.storage_variables_ordered,
                False,
                contract.custom_storage_layout,
            )
            self._compute_storage_layout(
                contract.name, contract.transient_variables_ordered, True, None
            )

    def _compute_storage_layout(
        self,
        contract_name: str,
        state_variables_ordered: list[StateVariable],
        is_transient: bool,
        custom_storage_layout: int | None,
    ):
        if is_transient:
            slot = 0
            self._transient_storage_layouts[contract_name] = {}
        else:
            slot = custom_storage_layout if custom_storage_layout else 0
            self._persistent_storage_layouts[contract_name] = {}

        offset = 0
        for var in state_variables_ordered:
            assert var.type
            size, new_slot = var.type.storage_size

            if new_slot:
                if offset > 0:
                    slot += 1
                    offset = 0
            elif size + offset > 32:
                slot += 1
                offset = 0

            if is_transient:
                self._transient_storage_layouts[contract_name][var.canonical_name] = (
                    slot,
                    offset,
                )
            else:
                self._persistent_storage_layouts[contract_name][var.canonical_name] = (
                    slot,
                    offset,
                )

            if new_slot:
                slot += math.ceil(size / 32)
            else:
                offset += size

    def storage_layout_of(self, contract: Contract, var: StateVariable) -> tuple[int, int]:
        if var.is_stored:
            return self._persistent_storage_layouts[contract.name][var.canonical_name]
        return self._transient_storage_layouts[contract.name][var.canonical_name]

    # endregion



import argparse
import logging
from pathlib import Path

from crytic_compile import cryticparser

from slither import Slither
from slither.utils.code_generation import generate_interface

logging.basicConfig()
logger = logging.getLogger("Slither-Interface")
logger.setLevel(logging.INFO)


def parse_args() -> argparse.Namespace:
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser = argparse.ArgumentParser(
        description="Generates code for a Solidity interface from contract",
        usage=("slither-interface <ContractName> <source file or deployment address>"),
    )

    parser.add_argument(
        "contract_source",
        help="The name of the contract (case sensitive) followed by the deployed contract address if verified on etherscan or project directory/filename for local contracts.",
        nargs="+",
    )

    parser.add_argument(
        "--unroll-structs",
        help="Whether to use structures' underlying types instead of the user-defined type",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--exclude-events",
        help="Excludes event signatures in the interface",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--exclude-errors",
        help="Excludes custom error signatures in the interface",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--exclude-enums",
        help="Excludes enum definitions in the interface",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--exclude-structs",
        help="Exclude struct definitions in the interface",
        default=False,
        action="store_true",
    )

    cryticparser.init(parser)

    return parser.parse_args()


def main() -> None:
    args = parse_args()

    contract_name, target = args.contract_source
    slither = Slither(target, **vars(args))

    _contract = slither.get_contract_from_name(contract_name)[0]

    interface = generate_interface(
        contract=_contract,
        unroll_structs=args.unroll_structs,
        include_events=not args.exclude_events,
        include_errors=not args.exclude_errors,
        include_enums=not args.exclude_enums,
        include_structs=not args.exclude_structs,
    )

    # add version pragma
    if _contract.compilation_unit.pragma_directives:
        interface = (
            f"pragma solidity {_contract.compilation_unit.pragma_directives[0].version};\n\n"
            + interface
        )

    # write interface to file
    export = Path("crytic-export", "interfaces")
    export.mkdir(parents=True, exist_ok=True)
    filename = f"I{contract_name}.sol"
    path = Path(export, filename)
    logger.info(f" Interface exported to {path}")
    with open(path, "w", encoding="utf8") as f:
        f.write(interface)


if __name__ == "__main__":
    main()



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class MultipleConstructorSchemes(AbstractDetector):
    """
    Module detecting multiple constructors in the same contract.
    (This was possible prior to Solidity 0.4.23, using old and new constructor schemes).
    """

    ARGUMENT = "multiple-constructors"
    HELP = "Multiple constructor schemes"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#multiple-constructor-schemes"
    )

    WIKI_TITLE = "Multiple constructor schemes"
    WIKI_DESCRIPTION = (
        "Detect multiple constructor definitions in the same contract (using new and old schemes)."
    )

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract A {
    uint x;
    constructor() public {
        x = 0;
    }
    function A() public {
        x = 1;
    }

    function test() public returns(uint) {
        return x;
    }
}
```
In Solidity [0.4.22](https://github.com/ethereum/solidity/releases/tag/v0.4.23), a contract with both constructor schemes will compile. The first constructor will take precedence over the second, which may be unintended."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Only declare one constructor, preferably using the new scheme `constructor(...)` instead of `function <contractName>(...)`."

    def _detect(self) -> list[Output]:
        """
        Detect multiple constructor schemes in the same contract
        :return: Returns a list of contract JSON result, where each result contains all constructor definitions.
        """
        results = []
        for contract in self.contracts:
            # Obtain any constructors defined in this contract
            constructors = [f for f in contract.constructors if f.contract_declarer == contract]

            # If there is more than one, we encountered the described issue occurring.
            if constructors and len(constructors) > 1:
                info: DETECTOR_INFO = [
                    contract,
                    " contains multiple constructors in the same contract:\n",
                ]
                for constructor in constructors:
                    info += ["\t- ", constructor, "\n"]

                res = self.generate_result(info)
                results.append(res)

        return results



"""
Check that the same pragma is used in all the files
"""

from collections import OrderedDict

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.formatters.attributes.constant_pragma import custom_format
from slither.utils.output import Output


class ConstantPragma(AbstractDetector):
    """
    Check that the same pragma is used in all the files
    """

    ARGUMENT = "pragma"
    HELP = "If different pragma directives are used"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#different-pragma-directives-are-used"

    WIKI_TITLE = "Different pragma directives are used"
    WIKI_DESCRIPTION = "Detect whether different Solidity versions are used."
    WIKI_RECOMMENDATION = "Use one Solidity version."

    def _detect(self) -> list[Output]:
        results = []
        pragma_directives_by_version = OrderedDict()
        for pragma in self.compilation_unit.pragma_directives:
            if pragma.is_solidity_version:
                if pragma.version not in pragma_directives_by_version:
                    pragma_directives_by_version[pragma.version] = [pragma]
                else:
                    pragma_directives_by_version[pragma.version].append(pragma)

        versions = list(pragma_directives_by_version.keys())
        if len(versions) > 1:
            info: DETECTOR_INFO = [f"{len(versions)} different versions of Solidity are used:\n"]

            for version in versions:
                pragmas = pragma_directives_by_version[version]
                info += [f"\t- Version constraint {version} is used by:\n"]
                for pragma in pragmas:
                    info += ["\t\t-", pragma, "\n"]

            res = self.generate_result(info)

            results.append(res)

        return results

    @staticmethod
    def _format(slither: SlitherCompilationUnit, result: dict) -> None:
        custom_format(slither, result)



from slither.core.solidity_types import FunctionType
from slither.core.variables.variable import Variable
from slither.slithir.operations.call import Call
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue
from slither.core.variables.local_variable import LocalVariable
from slither.slithir.variables.constant import Constant
from slither.slithir.variables.local_variable import LocalIRVariable
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA


class InternalDynamicCall(Call, OperationWithLValue):
    def __init__(
        self,
        lvalue: TemporaryVariableSSA | TemporaryVariable | None,
        function: LocalVariable | LocalIRVariable,
        function_type: FunctionType,
    ) -> None:
        assert isinstance(function_type, FunctionType)
        assert isinstance(function, Variable)
        assert is_valid_lvalue(lvalue) or lvalue is None
        super().__init__()
        self._function: Variable = function
        self._function_type = function_type
        self._lvalue = lvalue

        self._callid = None  # only used if gas/value != 0
        self._call_value = None
        self._call_gas = None

    @property
    def read(self) -> list[Constant | LocalIRVariable | LocalVariable]:
        return self._unroll(self.arguments) + [self.function]

    @property
    def function(self) -> Variable:
        return self._function

    @property
    def function_type(self) -> FunctionType:
        return self._function_type

    @property
    def call_value(self):
        return self._call_value

    @call_value.setter
    def call_value(self, v):
        self._call_value = v

    @property
    def call_gas(self):
        return self._call_gas

    @call_gas.setter
    def call_gas(self, v):
        self._call_gas = v

    @property
    def call_id(self):
        return self._callid

    @call_id.setter
    def call_id(self, c):
        self._callid = c

    def __str__(self):
        value = ""
        gas = ""
        args = [str(a) for a in self.arguments]
        if self.call_value:
            value = f"value:{self.call_value}"
        if self.call_gas:
            gas = f"gas:{self.call_gas}"
        if not self.lvalue:
            lvalue = ""
        elif isinstance(self.lvalue.type, (list,)):
            lvalue = f"{self.lvalue}({','.join(str(x) for x in self.lvalue.type)}) = "
        else:
            lvalue = f"{self.lvalue}({self.lvalue.type}) = "
        txt = "{}INTERNAL_DYNAMIC_CALL {}({}) {} {}"
        return txt.format(lvalue, self.function.name, ",".join(args), value, gas)



from slither.core.declarations import Contract, Function, Event
from slither.core.declarations.custom_error import CustomError
from slither.core.declarations.enum import Enum
from slither.core.declarations.solidity_import_placeholder import SolidityImportPlaceHolder
from slither.core.solidity_types import ElementaryType
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_rvalue, RVALUE
from slither.slithir.variables.constant import Constant
from slither.slithir.variables.reference import ReferenceVariable
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.slithir.variables.reference_ssa import ReferenceVariableSSA


class Member(OperationWithLValue):
    def __init__(
        self,
        variable_left: SourceMapping,
        variable_right: Constant,
        result: ReferenceVariable | ReferenceVariableSSA,
    ) -> None:
        # Function can happen for something like
        # library FunctionExtensions {
        #     function h(function() internal _t, uint8) internal {  }
        # }
        # contract FunctionMembers {
        #     using FunctionExtensions for function();
        #
        #     function f() public {
        #         f.h(1);
        #     }
        # }
        # Can be an ElementaryType because of bytes.concat, string.concat
        assert is_valid_rvalue(variable_left) or isinstance(
            variable_left,
            (
                Contract,
                Enum,
                Function,
                Event,
                CustomError,
                SolidityImportPlaceHolder,
                ElementaryType,
            ),
        )

        assert isinstance(variable_right, Constant)
        assert isinstance(result, ReferenceVariable)
        super().__init__()
        self._variable_left: (
            RVALUE
            | Contract
            | Enum
            | Function
            | Event
            | CustomError
            | SolidityImportPlaceHolder
            | ElementaryType
        ) = variable_left
        self._variable_right = variable_right
        self._lvalue = result
        self._gas = None
        self._value = None

    @property
    def read(self) -> list[SourceMapping]:
        return [self.variable_left, self.variable_right]

    @property
    def variable_left(
        self,
    ) -> (
        RVALUE
        | Contract
        | Enum
        | Function
        | CustomError
        | SolidityImportPlaceHolder
        | ElementaryType
    ):
        return self._variable_left

    @property
    def variable_right(self) -> Constant:
        return self._variable_right

    @property
    def call_value(self):
        return self._value

    @call_value.setter
    def call_value(self, v):
        self._value = v

    @property
    def call_gas(self):
        return self._gas

    @call_gas.setter
    def call_gas(self, gas):
        self._gas = gas

    def __str__(self):
        return f"{self.lvalue}({self.lvalue.type}) -> {self.variable_left}.{self.variable_right}"



from pathlib import Path
import pytest

from crytic_compile import CryticCompile
from crytic_compile.platform.solc_standard_json import SolcStandardJson

from slither import Slither

from tests.utils import _run_all_detectors


TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"

hardhat_available = Path(TEST_DATA_DIR, "test_node_modules/node_modules/hardhat").exists()


@pytest.mark.skipif(not hardhat_available, reason="requires Hardhat and project setup")
def test_node_modules() -> None:
    # hardhat must have been installed in tests/test_node_modules
    # For the CI its done through the github action config

    slither = Slither(Path(TEST_DATA_DIR, "test_node_modules").as_posix())
    _run_all_detectors(slither)


def test_contract_name_collision(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    standard_json = SolcStandardJson()
    standard_json.add_source_file(
        Path(TEST_DATA_DIR, "test_contract_name_collisions", "a.sol").as_posix()
    )
    standard_json.add_source_file(
        Path(TEST_DATA_DIR, "test_contract_name_collisions", "b.sol").as_posix()
    )

    compilation = CryticCompile(standard_json, solc=solc_path)
    slither = Slither(compilation)

    _run_all_detectors(slither)


def test_cycle(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    slither = Slither(Path(TEST_DATA_DIR, "test_cyclic_import", "a.sol").as_posix(), solc=solc_path)
    _run_all_detectors(slither)


def test_contract_function_parameter(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    standard_json = SolcStandardJson()
    standard_json.add_source_file(
        Path(TEST_DATA_DIR, "test_contract_data", "test_contract_data.sol").as_posix()
    )
    compilation = CryticCompile(standard_json, solc=solc_path)
    slither = Slither(compilation)
    contract = slither.contracts[0]
    function = contract.functions[0]
    parameters = function.parameters

    assert parameters[0].name == "param1"
    assert parameters[1].name == ""
    assert parameters[2].name == "param3"



from slither.core.source_mapping.source_mapping import SourceMapping


class TopLevel(SourceMapping):
    """
    This class is used to represent objects that are at the top level
    The opposite is ContractLevel

    """



# Functions for generating Solidity code
from typing import TYPE_CHECKING

from slither.utils.type import (
    convert_type_for_solidity_signature_to_string,
    export_nested_types_from_variable,
    export_return_type_from_variable,
)
from slither.core.solidity_types import (
    Type,
    UserDefinedType,
    MappingType,
    ArrayType,
    ElementaryType,
    TypeAlias,
)
from slither.core.declarations import Structure, StructureContract, Enum, Contract

if TYPE_CHECKING:
    from slither.core.declarations import FunctionContract, CustomErrorContract
    from slither.core.variables.state_variable import StateVariable
    from slither.core.variables.local_variable import LocalVariable


def generate_interface(
    contract: "Contract",
    unroll_structs: bool = True,
    include_events: bool = True,
    include_errors: bool = True,
    include_enums: bool = True,
    include_structs: bool = True,
) -> str:
    """
    Generates code for a Solidity interface to the contract.
    Args:
        contract: A Contract object.
        unroll_structs: Whether to use structures' underlying types instead of the user-defined type (default: True).
        include_events: Whether to include event signatures in the interface (default: True).
        include_errors: Whether to include custom error signatures in the interface (default: True).
        include_enums: Whether to include enum definitions in the interface (default: True).
        include_structs: Whether to include struct definitions in the interface (default: True).

    Returns:
        A string with the code for an interface, with function stubs for all public or external functions and
        state variables, as well as any events, custom errors and/or structs declared in the contract.
    """
    interface = f"interface I{contract.name} {{\n"
    if include_events:
        for event in contract.events:
            name, args = event.signature
            interface += f"    event {name}({', '.join(args)});\n"
    if include_errors:
        for error in contract.custom_errors:
            interface += f"    error {generate_custom_error_interface(error, unroll_structs)};\n"
    if include_enums:
        for enum in contract.enums:
            interface += f"    enum {enum.name} {{ {', '.join(enum.values)} }}\n"
    if include_structs:
        # Include structures defined in this contract and at the top level
        structs = contract.structures + contract.compilation_unit.structures_top_level
        # Function signatures may reference other structures as well
        # Include structures defined in libraries used for them
        for _for in contract.using_for:
            if (
                isinstance(_for, UserDefinedType)
                and isinstance(_for.type, StructureContract)
                and _for.type not in structs
            ):
                structs.append(_for.type)
        # Include any other structures used as function arguments/returns
        for func in contract.functions_entry_points:
            for arg in func.parameters + func.returns:
                _type = arg.type
                if isinstance(_type, ArrayType):
                    _type = _type.type
                while isinstance(_type, MappingType):
                    _type = _type.type_to
                if isinstance(_type, UserDefinedType):
                    _type = _type.type
                if isinstance(_type, Structure) and _type not in structs:
                    structs.append(_type)
        for struct in structs:
            interface += generate_struct_interface_str(struct, indent=4)
            for elem in struct.elems_ordered:
                if (
                    isinstance(elem.type, UserDefinedType)
                    and isinstance(elem.type.type, StructureContract)
                    and elem.type.type not in structs
                ):
                    structs.append(elem.type.type)
    for var in contract.state_variables_entry_points:
        # if any(func.name == var.name for func in contract.functions_entry_points):
        #     # ignore public variables that override a public function
        #     continue
        var_sig = generate_interface_variable_signature(var, unroll_structs)
        if var_sig is not None and var_sig != "":
            interface += f"    function {var_sig};\n"
    for func in contract.functions_entry_points:
        if func.is_constructor or func.is_fallback or func.is_receive or not func.is_implemented:
            continue
        interface += (
            f"    function {generate_interface_function_signature(func, unroll_structs)};\n"
        )
    interface += "}\n\n"
    return interface


def generate_interface_variable_signature(
    var: "StateVariable", unroll_structs: bool = True
) -> str | None:
    if var.visibility in ["private", "internal"]:
        return None
    if isinstance(var.type, UserDefinedType) and isinstance(var.type.type, Structure):
        for elem in var.type.type.elems_ordered:
            if isinstance(elem.type, MappingType):
                return ""
    if unroll_structs:
        params = [
            convert_type_for_solidity_signature_to_string(x).replace("(", "").replace(")", "")
            for x in export_nested_types_from_variable(var)
        ]
        returns = [
            convert_type_for_solidity_signature_to_string(x).replace("(", "").replace(")", "")
            for x in export_return_type_from_variable(var)
        ]
    else:
        _, params, _ = var.signature
        params = [p + " memory" if p in ["bytes", "string"] else p for p in params]
        returns = []
        _type = var.type
        while isinstance(_type, MappingType):
            _type = _type.type_to
        while isinstance(_type, (ArrayType, UserDefinedType)):
            _type = _type.type
        if isinstance(_type, TypeAlias):
            _type = _type.type
        if isinstance(_type, Structure):
            if any(isinstance(elem.type, MappingType) for elem in _type.elems_ordered):
                return ""
        ret = str(_type)
        if isinstance(_type, Structure) or (isinstance(_type, Type) and _type.is_dynamic):
            ret += " memory"
        elif isinstance(_type, Contract):
            ret = "address"
        returns.append(ret)
    return f"{var.name}({','.join(params)}) external returns ({', '.join(returns)})"


def generate_interface_function_signature(
    func: "FunctionContract", unroll_structs: bool = True
) -> str | None:
    """
    Generates a string of the form:
        func_name(type1,type2) external {payable/view/pure} returns (type3)

    Args:
        func: A FunctionContract object
        unroll_structs: Determines whether structs are unrolled into underlying types (default: True)

    Returns:
        The function interface as a str (contains the return values).
        Returns None if the function is private or internal, or is a constructor/fallback/receive.
    """

    def format_var(var: "LocalVariable", unroll: bool) -> str:
        if unroll:
            return (
                convert_type_for_solidity_signature_to_string(var.type)
                .replace("(", "")
                .replace(")", "")
            )
        if var.type.is_dynamic:
            return f"{_handle_dynamic_struct_elem(var.type)} {var.location}"
        if isinstance(var.type, ArrayType) and isinstance(
            var.type.type, (UserDefinedType, ElementaryType)
        ):
            return (
                convert_type_for_solidity_signature_to_string(var.type)
                .replace("(", "")
                .replace(")", "")
                + f" {var.location}"
            )
        if isinstance(var.type, UserDefinedType):
            if isinstance(var.type.type, Structure):
                return f"{var.type.type!s} memory"
            if isinstance(var.type.type, Enum):
                return str(var.type.type)
            if isinstance(var.type.type, Contract):
                return "address"
        if isinstance(var.type, TypeAlias):
            return str(var.type.type)
        return str(var.type)

    name, _, _ = func.signature
    if (
        func not in func.contract.functions_entry_points
        or func.is_constructor
        or func.is_fallback
        or func.is_receive
    ):
        return None
    view = " view" if func.view and not func.pure else ""
    pure = " pure" if func.pure else ""
    payable = " payable" if func.payable else ""
    # Make sure the function doesn't return a struct with nested mappings
    for ret in func.returns:
        if isinstance(ret.type, UserDefinedType) and isinstance(ret.type.type, Structure):
            for elem in ret.type.type.elems_ordered:
                if isinstance(elem.type, MappingType):
                    return ""
    returns = [format_var(ret, unroll_structs) for ret in func.returns]
    parameters = [format_var(param, unroll_structs) for param in func.parameters]
    _interface_signature_str = (
        name + "(" + ",".join(parameters) + ") external" + payable + pure + view
    )
    if len(returns) > 0:
        _interface_signature_str += " returns (" + ",".join(returns) + ")"
    return _interface_signature_str


def generate_struct_interface_str(struct: "Structure", indent: int = 0) -> str:
    """
    Generates code for a structure declaration in an interface of the form:
        struct struct_name {
            elem1_type elem1_name;
            elem2_type elem2_name;
            ...        ...
        }
    Args:
        struct: A Structure object.
        indent: Number of spaces to indent the code block with.

    Returns:
        The structure declaration code as a string.
    """
    spaces = ""
    for _ in range(0, indent):
        spaces += " "
    definition = f"{spaces}struct {struct.name} {{\n"
    for elem in struct.elems_ordered:
        if elem.type.is_dynamic:
            definition += f"{spaces}    {_handle_dynamic_struct_elem(elem.type)} {elem.name};\n"
        elif isinstance(elem.type, UserDefinedType):
            if isinstance(elem.type.type, Structure):
                definition += f"{spaces}    {elem.type.type} {elem.name};\n"
            else:
                definition += f"{spaces}    {convert_type_for_solidity_signature_to_string(elem.type)} {elem.name};\n"
        elif isinstance(elem.type, TypeAlias):
            definition += f"{spaces}    {elem.type.type} {elem.name};\n"
        else:
            definition += f"{spaces}    {elem.type} {elem.name};\n"
    definition += f"{spaces}}}\n"
    return definition


def _handle_dynamic_struct_elem(elem_type: Type) -> str:
    assert elem_type.is_dynamic
    if isinstance(elem_type, ElementaryType):
        return f"{elem_type}"
    if isinstance(elem_type, ArrayType):
        base_type = elem_type.type
        if isinstance(base_type, UserDefinedType):
            if isinstance(base_type.type, Contract):
                return "address[]"
            if isinstance(base_type.type, Enum):
                return convert_type_for_solidity_signature_to_string(elem_type)
            return f"{base_type.type.name}[]"
        return f"{base_type}[]"
    if isinstance(elem_type, MappingType):
        type_to = elem_type.type_to
        type_from = elem_type.type_from
        if isinstance(type_from, UserDefinedType) and isinstance(type_from.type, Contract):
            type_from = ElementaryType("address")
        if isinstance(type_to, MappingType):
            return f"mapping({type_from} => {_handle_dynamic_struct_elem(type_to)})"
        if isinstance(type_to, UserDefinedType):
            if isinstance(type_to.type, Contract):
                return f"mapping({type_from} => address)"
            return f"mapping({type_from} => {type_to.type.name})"
        return f"{elem_type}"
    return ""


def generate_custom_error_interface(
    error: "CustomErrorContract", unroll_structs: bool = True
) -> str:
    args = [
        convert_type_for_solidity_signature_to_string(arg.type).replace("(", "").replace(")", "")
        if unroll_structs
        else str(arg.type.type)
        if isinstance(arg.type, UserDefinedType) and isinstance(arg.type.type, (Structure, Enum))
        else str(arg.type)
        for arg in error.parameters
    ]
    return f"{error.name}({', '.join(args)})"



from collections.abc import Callable
from slither.vyper_parsing.ast.types import (
    ASTNode,
    Module,
    ImportFrom,
    EventDef,
    AnnAssign,
    Name,
    Call,
    StructDef,
    VariableDecl,
    Subscript,
    Index,
    Hex,
    Int,
    Str,
    Tuple,
    FunctionDef,
    Assign,
    Raise,
    Attribute,
    Assert,
    Keyword,
    Arguments,
    Arg,
    UnaryOp,
    BinOp,
    Expr,
    Log,
    Return,
    VyDict,
    VyList,
    NameConstant,
    If,
    Compare,
    For,
    Break,
    Continue,
    Pass,
    InterfaceDef,
    EnumDef,
    Bytes,
    AugAssign,
    BoolOp,
)


class ParsingError(Exception):
    pass


def _extract_base_props(raw: dict) -> dict:
    return {
        "src": raw["src"],
        "node_id": raw["node_id"],
    }


def _extract_decl_props(raw: dict) -> dict:
    return {
        "doc_string": parse_doc_str(raw["doc_string"]) if raw["doc_string"] else None,
        **_extract_base_props(raw),
    }


def parse_module(raw: dict) -> Module:
    nodes_parsed: list[ASTNode] = []

    for node in raw["body"]:
        nodes_parsed.append(parse(node))

    return Module(name=raw["name"], body=nodes_parsed, **_extract_decl_props(raw))


def parse_import_from(raw: dict) -> ImportFrom:
    return ImportFrom(
        module=raw["module"],
        name=raw["name"],
        alias=raw["alias"],
        **_extract_base_props(raw),
    )


def parse_event_def(raw: dict) -> EventDef:
    body_parsed: list[ASTNode] = []
    for node in raw["body"]:
        body_parsed.append(parse(node))

    return EventDef(
        name=raw["name"],
        body=body_parsed,
        *_extract_base_props(raw),
    )


def parse_ann_assign(raw: dict) -> AnnAssign:
    return AnnAssign(
        target=parse(raw["target"]),
        annotation=parse(raw["annotation"]),
        value=parse(raw["value"]) if raw["value"] else None,
        **_extract_base_props(raw),
    )


def parse_name(raw: dict) -> Name:
    return Name(
        id=raw["id"],
        **_extract_base_props(raw),
    )


def parse_call(raw: dict) -> Call:
    return Call(
        func=parse(raw["func"]),
        args=[parse(arg) for arg in raw["args"]],
        keyword=parse(raw["keyword"]) if raw["keyword"] else None,
        keywords=[parse(keyword) for keyword in raw["keywords"]],
        **_extract_base_props(raw),
    )


def parse_struct_def(raw: dict) -> StructDef:
    body_parsed: list[ASTNode] = []
    for node in raw["body"]:
        body_parsed.append(parse(node))

    return StructDef(
        name=raw["name"],
        body=body_parsed,
        **_extract_base_props(raw),
    )


def parse_variable_decl(raw: dict) -> VariableDecl:
    return VariableDecl(
        annotation=parse(raw["annotation"]),
        value=parse(raw["value"]) if raw["value"] else None,
        target=parse(raw["target"]),
        is_constant=raw["is_constant"],
        is_immutable=raw["is_immutable"],
        is_public=raw["is_public"],
        **_extract_base_props(raw),
    )


def parse_subscript(raw: dict) -> Subscript:
    return Subscript(
        value=parse(raw["value"]),
        slice=parse(raw["slice"]),
        **_extract_base_props(raw),
    )


def parse_index(raw: dict) -> Index:
    return Index(value=parse(raw["value"]), **_extract_base_props(raw))


def parse_bytes(raw: dict) -> Bytes:
    return Bytes(value=raw["value"], **_extract_base_props(raw))


def parse_hex(raw: dict) -> Hex:
    return Hex(value=raw["value"], **_extract_base_props(raw))


def parse_int(raw: dict) -> Int:
    return Int(value=raw["value"], **_extract_base_props(raw))


def parse_str(raw: dict) -> Str:
    return Str(value=raw["value"], **_extract_base_props(raw))


def parse_tuple(raw: dict) -> ASTNode:
    return Tuple(elements=[parse(elem) for elem in raw["elements"]], **_extract_base_props(raw))


def parse_function_def(raw: dict) -> FunctionDef:
    body_parsed: list[ASTNode] = []
    for node in raw["body"]:
        body_parsed.append(parse(node))

    decorators_parsed: list[ASTNode] = []
    for node in raw["decorator_list"]:
        decorators_parsed.append(parse(node))

    return FunctionDef(
        name=raw["name"],
        args=parse_arguments(raw["args"]),
        returns=parse(raw["returns"]) if raw["returns"] else None,
        body=body_parsed,
        pos=raw["pos"],
        decorators=decorators_parsed,
        **_extract_decl_props(raw),
    )


def parse_assign(raw: dict) -> Assign:
    return Assign(
        target=parse(raw["target"]),
        value=parse(raw["value"]),
        **_extract_base_props(raw),
    )


def parse_attribute(raw: dict) -> Attribute:
    return Attribute(
        value=parse(raw["value"]),
        attr=raw["attr"],
        **_extract_base_props(raw),
    )


def parse_arguments(raw: dict) -> Arguments:
    return Arguments(
        args=[parse_arg(arg) for arg in raw["args"]],
        default=parse(raw["default"]) if raw["default"] else None,
        defaults=[parse(x) for x in raw["defaults"]],
        **_extract_base_props(raw),
    )


def parse_arg(raw: dict) -> Arg:
    return Arg(arg=raw["arg"], annotation=parse(raw["annotation"]), **_extract_base_props(raw))


def parse_assert(raw: dict) -> Assert:
    return Assert(
        test=parse(raw["test"]),
        msg=parse(raw["msg"]) if raw["msg"] else None,
        **_extract_base_props(raw),
    )


def parse_raise(raw: dict) -> Raise:
    return Raise(exc=parse(raw["exc"]) if raw["exc"] else None, **_extract_base_props(raw))


def parse_expr(raw: dict) -> Expr:
    return Expr(value=parse(raw["value"]), **_extract_base_props(raw))


# This is done for convenience so we can call `UnaryOperationType.get_type` during expression parsing.
unop_ast_type_to_op_symbol = {"Not": "!", "USub": "-"}


def parse_unary_op(raw: dict) -> UnaryOp:
    unop_str = unop_ast_type_to_op_symbol[raw["op"]["ast_type"]]
    return UnaryOp(op=unop_str, operand=parse(raw["operand"]), **_extract_base_props(raw))


# This is done for convenience so we can call `BinaryOperationType.get_type` during expression parsing.
binop_ast_type_to_op_symbol = {
    "Add": "+",
    "Mult": "*",
    "Sub": "-",
    "Div": "/",
    "Pow": "**",
    "Mod": "%",
    "BitAnd": "&",
    "BitOr": "|",
    "Shr": "<<",
    "Shl": ">>",
    "NotEq": "!=",
    "Eq": "==",
    "LtE": "<=",
    "GtE": ">=",
    "Lt": "<",
    "Gt": ">",
    "In": "In",
    "NotIn": "NotIn",
}


def parse_bin_op(raw: dict) -> BinOp:
    arith_op_str = binop_ast_type_to_op_symbol[raw["op"]["ast_type"]]
    return BinOp(
        left=parse(raw["left"]),
        op=arith_op_str,
        right=parse(raw["right"]),
        **_extract_base_props(raw),
    )


def parse_compare(raw: dict) -> Compare:
    logical_op_str = binop_ast_type_to_op_symbol[raw["op"]["ast_type"]]
    return Compare(
        left=parse(raw["left"]),
        op=logical_op_str,
        right=parse(raw["right"]),
        **_extract_base_props(raw),
    )


def parse_keyword(raw: dict) -> Keyword:
    return Keyword(arg=raw["arg"], value=parse(raw["value"]), **_extract_base_props(raw))


def parse_log(raw: dict) -> Log:
    return Log(value=parse(raw["value"]), **_extract_base_props(raw))


def parse_return(raw: dict) -> Return:
    return Return(value=parse(raw["value"]) if raw["value"] else None, **_extract_base_props(raw))


def parse_dict(raw: dict) -> ASTNode:
    return VyDict(
        keys=[parse(x) for x in raw["keys"]],
        values=[parse(x) for x in raw["values"]],
        **_extract_base_props(raw),
    )


def parse_list(raw: dict) -> VyList:
    return VyList(elements=[parse(x) for x in raw["elements"]], **_extract_base_props(raw))


def parse_name_constant(raw: dict) -> NameConstant:
    return NameConstant(value=raw["value"], **_extract_base_props(raw))


def parse_doc_str(raw: dict) -> str:
    assert isinstance(raw["value"], str)
    return raw["value"]


def parse_if(raw: dict) -> ASTNode:
    return If(
        test=parse(raw["test"]),
        body=[parse(x) for x in raw["body"]],
        orelse=[parse(x) for x in raw["orelse"]],
        **_extract_base_props(raw),
    )


def parse_for(raw: dict) -> For:
    return For(
        target=parse(raw["target"]),
        iter=parse(raw["iter"]),
        body=[parse(x) for x in raw["body"]],
        **_extract_base_props(raw),
    )


def parse_break(raw: dict) -> Break:
    return Break(**_extract_base_props(raw))


def parse_continue(raw: dict) -> Continue:
    return Continue(**_extract_base_props(raw))


def parse_pass(raw: dict) -> Pass:
    return Pass(
        **_extract_base_props(raw),
    )


def parse_interface_def(raw: dict) -> InterfaceDef:
    nodes_parsed: list[ASTNode] = []

    for node in raw["body"]:
        nodes_parsed.append(parse(node))
    return InterfaceDef(name=raw["name"], body=nodes_parsed, **_extract_base_props(raw))


def parse_enum_def(raw: dict) -> EnumDef:
    nodes_parsed: list[ASTNode] = []

    for node in raw["body"]:
        nodes_parsed.append(parse(node))

    return EnumDef(name=raw["name"], body=nodes_parsed, **_extract_base_props(raw))


aug_assign_ast_type_to_op_symbol = {
    "Add": "+=",
    "Mult": "*=",
    "Sub": "-=",
    "Div": "-=",
    "Pow": "**=",
    "Mod": "%=",
    "BitAnd": "&=",
    "BitOr": "|=",
    "Shr": "<<=",
    "Shl": ">>=",
}


def parse_aug_assign(raw: dict) -> AugAssign:
    op_str = aug_assign_ast_type_to_op_symbol[raw["op"]["ast_type"]]
    return AugAssign(
        target=parse(raw["target"]),
        op=op_str,
        value=parse(raw["value"]),
        **_extract_base_props(raw),
    )


def parse_unsupported(raw: dict) -> ASTNode:
    raise ParsingError("unsupported Vyper node", raw["ast_type"], raw.keys(), raw)


bool_op_ast_type_to_op_symbol = {"And": "&&", "Or": "||"}


def parse_bool_op(raw: dict) -> BoolOp:
    op_str = bool_op_ast_type_to_op_symbol[raw["op"]["ast_type"]]
    return BoolOp(op=op_str, values=[parse(x) for x in raw["values"]], **_extract_base_props(raw))


def parse(raw: dict) -> ASTNode:
    try:
        return PARSERS.get(raw["ast_type"], parse_unsupported)(raw)
    except ParsingError as e:
        raise e
    except Exception as e:
        raise e
        # raise ParsingError("failed to parse Vyper node", raw["ast_type"], e, raw.keys(), raw)


PARSERS: dict[str, Callable[[dict], ASTNode]] = {
    "Module": parse_module,
    "ImportFrom": parse_import_from,
    "EventDef": parse_event_def,
    "AnnAssign": parse_ann_assign,
    "Name": parse_name,
    "Call": parse_call,
    "Pass": parse_pass,
    "StructDef": parse_struct_def,
    "VariableDecl": parse_variable_decl,
    "Subscript": parse_subscript,
    "Index": parse_index,
    "Hex": parse_hex,
    "Int": parse_int,
    "Str": parse_str,
    "DocStr": parse_doc_str,
    "Tuple": parse_tuple,
    "FunctionDef": parse_function_def,
    "Assign": parse_assign,
    "Raise": parse_raise,
    "Attribute": parse_attribute,
    "Assert": parse_assert,
    "keyword": parse_keyword,
    "arguments": parse_arguments,
    "arg": parse_arg,
    "UnaryOp": parse_unary_op,
    "BinOp": parse_bin_op,
    "Expr": parse_expr,
    "Log": parse_log,
    "Return": parse_return,
    "If": parse_if,
    "Dict": parse_dict,
    "List": parse_list,
    "Compare": parse_compare,
    "NameConstant": parse_name_constant,
    "For": parse_for,
    "Break": parse_break,
    "Continue": parse_continue,
    "InterfaceDef": parse_interface_def,
    "EnumDef": parse_enum_def,
    "Bytes": parse_bytes,
    "AugAssign": parse_aug_assign,
    "BoolOp": parse_bool_op,
}



import itertools
from typing import Any

from slither.core.solidity_types.type import Type


# see https://solidity.readthedocs.io/en/v0.4.24/miscellaneous.html?highlight=grammar
from slither.exceptions import SlitherException

Int = [
    "int",
    "int8",
    "int16",
    "int24",
    "int32",
    "int40",
    "int48",
    "int56",
    "int64",
    "int72",
    "int80",
    "int88",
    "int96",
    "int104",
    "int112",
    "int120",
    "int128",
    "int136",
    "int144",
    "int152",
    "int160",
    "int168",
    "int176",
    "int184",
    "int192",
    "int200",
    "int208",
    "int216",
    "int224",
    "int232",
    "int240",
    "int248",
    "int256",
]

Max_Int = {k: 2 ** (8 * i - 1) - 1 if i > 0 else 2**255 - 1 for i, k in enumerate(Int)}
Min_Int = {k: -(2 ** (8 * i - 1)) if i > 0 else -(2**255) for i, k in enumerate(Int)}

Uint = [
    "uint",
    "uint8",
    "uint16",
    "uint24",
    "uint32",
    "uint40",
    "uint48",
    "uint56",
    "uint64",
    "uint72",
    "uint80",
    "uint88",
    "uint96",
    "uint104",
    "uint112",
    "uint120",
    "uint128",
    "uint136",
    "uint144",
    "uint152",
    "uint160",
    "uint168",
    "uint176",
    "uint184",
    "uint192",
    "uint200",
    "uint208",
    "uint216",
    "uint224",
    "uint232",
    "uint240",
    "uint248",
    "uint256",
]

Max_Uint = {k: 2 ** (8 * i) - 1 if i > 0 else 2**256 - 1 for i, k in enumerate(Uint)}
Min_Uint = dict.fromkeys(Uint, 0)


Byte = [
    "byte",
    "bytes",
    "bytes1",
    "bytes2",
    "bytes3",
    "bytes4",
    "bytes5",
    "bytes6",
    "bytes7",
    "bytes8",
    "bytes9",
    "bytes10",
    "bytes11",
    "bytes12",
    "bytes13",
    "bytes14",
    "bytes15",
    "bytes16",
    "bytes17",
    "bytes18",
    "bytes19",
    "bytes20",
    "bytes21",
    "bytes22",
    "bytes23",
    "bytes24",
    "bytes25",
    "bytes26",
    "bytes27",
    "bytes28",
    "bytes29",
    "bytes30",
    "bytes31",
    "bytes32",
]

Max_Byte = {k: 2 ** (8 * (i + 1)) - 1 for i, k in enumerate(Byte[2:])}
Max_Byte["bytes"] = None
Max_Byte["string"] = None
Max_Byte["byte"] = 255
Min_Byte = dict.fromkeys(Byte, 0)
Min_Byte["bytes"] = 0x0
Min_Byte["string"] = 0x0
Min_Byte["byte"] = 0x0

MaxValues = dict(dict(Max_Int, **Max_Uint), **Max_Byte)
MinValues = dict(dict(Min_Int, **Min_Uint), **Min_Byte)

# https://solidity.readthedocs.io/en/v0.4.24/types.html#fixed-point-numbers
M = list(range(8, 257, 8))
N = list(range(0, 81))
MN = list(itertools.product(M, N))

Fixed = [f"fixed{m}x{n}" for (m, n) in MN] + ["fixed"]
Ufixed = [f"ufixed{m}x{n}" for (m, n) in MN] + ["ufixed"]

ElementaryTypeName = ["address", "bool", "string", "var"] + Int + Uint + Byte + Fixed + Ufixed


class NonElementaryType(Exception):
    pass


class ElementaryType(Type):
    def __init__(self, t: str) -> None:
        if t not in ElementaryTypeName:
            raise NonElementaryType
        super().__init__()
        if t == "uint":
            t = "uint256"
        elif t == "int":
            t = "int256"
        elif t == "byte":
            t = "bytes1"
        self._type = t

    @property
    def is_dynamic(self) -> bool:
        return self._type in ("bytes", "string")

    @property
    def type(self) -> str:
        return self._type

    @property
    def name(self) -> str:
        return self.type

    @property
    def size(self) -> int | None:
        """
            Return the size in bits
            Return None if the size is not known
        Returns:
            int
        """
        t = self._type
        if t.startswith("uint"):
            return int(t[len("uint") :])
        if t.startswith("int"):
            return int(t[len("int") :])
        if t == "bool":
            return 8
        if t == "address":
            return 160
        if t.startswith("bytes") and t != "bytes":
            return int(t[len("bytes") :]) * 8
        raise SlitherException(f"{t} does not have a size")

    @property
    def storage_size(self) -> tuple[int, bool]:
        if self._type in ["string", "bytes"]:
            return 32, True
        if self.size is None:
            return 32, True
        return int(self.size / 8), False

    @property
    def min(self) -> int:
        if self.name in MinValues:
            return MinValues[self.name]
        raise SlitherException(f"{self.name} does not have a min value")

    @property
    def max(self) -> int:
        if self.name in MaxValues:
            return MaxValues[self.name]
        raise SlitherException(f"{self.name} does not have a max value")

    def __str__(self) -> str:
        return self._type

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ElementaryType):
            return False
        return self.type == other.type

    def __hash__(self) -> int:
        return hash(self._type)
